!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!     File lsqrblas.f90   (double precision)
!
!     This file contains the following BLAS routines
!        ddot, dnrm2, dscal
!     required by subroutines LSQR and Acheck.
!
! 24 Sep 2007: All parameters declared with correct intent
!              to avoid compiler warnings.
! 24 Oct 2007: Use real(8) instead of double precision or -r8.
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!! DDOT forms the dot product of two vectors.
!
!  Discussion:
!    This routine uses double precision real arithmetic.
!    This routine uses unrolled loops for increments equal to one.
!
!  Modified:
!    16 May 2005
!
!  Author:
!    Jack Dongarra
!    Fortran90 translation by John Burkardt.
!
!  Reference:
!    Jack Dongarra, Jim Bunch, Cleve Moler, Pete Stewart,
!    LINPACK User's Guide,
!    SIAM, 1979,
!    ISBN13: 978-0-898711-72-1,
!    LC: QA214.L56.
!
!    Charles Lawson, Richard Hanson, David Kincaid, Fred Krogh,
!    Algorithm 539, 
!    Basic Linear Algebra Subprograms for Fortran Usage,
!    ACM Transactions on Mathematical Software, 
!    Volume 5, Number 3, September 1979, pages 308-323.
!
!  Parameters:
!
!    Input, integer N, the number of entries in the vectors. Note: this is not the length of the vector
!
!    Input, real ( kind = dp ) DX(:), the first vector.
!
!    Input, integer INCX, the increment between successive entries in DX.
!
!    Input, real ( kind = dp ) DY(:), the second vector.
!
!    Input, integer INCY, the increment between successive entries in DY.
!
!    Output, real ( kind = dp ) DDOT, the sum of the product of the 
!    corresponding entries of DX and DY.

      ! In fact we should sum them from smallest to largest if possible
      ! This is not done for speed consideration
      real(kind=dp) function ddot(n,dx,incx,dy,incy)
      use lsqrDataModule
      implicit none
      integer, intent(in) :: n,incx,incy
      real(kind=dp), intent(in) :: dx(:),dy(:)

      real(kind=dp)  :: dtemp
      integer :: i,ix,iy

      dtemp = 0.0_dp
      ddot = 0.0_dp

      if (n <= 0) then
          return
      end if

      !  Code for unequal increments or equal increments
      !  not equal to 1.

      if (0 <= incx) then ! For positive increments
          if (1 + (n - 1) * incx > size(dx)) then
              write(*,'(a)') "Error: The first vector is too small."
              stop
          else
              ix = 1
          end if
      else
          if ((-n + 1) * incx + 1 > size(dx)) then
              write(*,'(a)') "Error: The first vector is too small."
              stop
          else
              ix = (-n + 1) * incx + 1 ! For negative increments
          end if
      end if

      if (0 <= incy) then
          if (1 + (n - 1) * incy > size(dy)) then
              write(*,'(a)') "Error: The second vector is too small."
              stop
          else
              iy = 1
          end if
      else
          if ((- n + 1) * incy + 1 > size(dy)) then
              write(*,'(a)') "Error: The second vector is too small."
              stop
          else
              iy = (- n + 1) * incy + 1
          end if
      end if

      do i = 1, n
          dtemp = dtemp + dx(ix) * dy(iy)
          ix = ix + incx
          iy = iy + incy
      end do

      ddot = dtemp
      return

end function ddot

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!*****************************************************************************
!
!! DNRM2 returns the euclidean norm of a vector.
!
!  Discussion:
!    This routine uses real(8) real arithmetic.
!     DNRM2 ( X ) = sqrt ( X' * X )
!
!  Modified:
!    16 May 2005
!
!  Author:
!    Sven Hammarling
!    Fortran90 translation by John Burkardt.
!
!  Reference:
!    Jack Dongarra, Jim Bunch, Cleve Moler, Pete Stewart,
!    LINPACK User's Guide,
!    SIAM, 1979,
!    ISBN13: 978-0-898711-72-1,
!    LC: QA214.L56.
!
!    Charles Lawson, Richard Hanson, David Kincaid, Fred Krogh,
!    Algorithm 539, 
!    Basic Linear Algebra Subprograms for Fortran Usage,
!    ACM Transactions on Mathematical Software,
!    Volume 5, Number 3, September 1979, pages 308-323.
!
!  Parameters:
!
!    Input, integer N, the number of entries in the vector.
!
!    Input, real ( kind = 8 ) X(*), the vector whose norm is to be computed.
!
!    Input, integer INCX, the increment between successive entries of X.
!
!    Output, real ( kind = 8 ) DNRM2, the Euclidean norm of X.
!

      real(kind=dp) function dnrm2(n, x, incx)
      use lsqrDataModule
      implicit none
      integer, intent(in) :: n,incx
      real(kind=dp), intent(in) :: x(:)

      integer :: i,ix
      real(kind=dp) :: ssq,absxi,scale

      scale = 0.0_dp
      ssq = 1.0_dp
      dnrm2  = 0.0_dp

      if (n <= 0) then
          return
      end if

      if (0 <= incx) then ! For positive increments
          if (1 + (n - 1) * incx > size(x)) then
              write(*,'(a)') "Error: The vector is too small."
              stop
          else
              ix = 1
          end if
      else ! For negative increments
          if ((- n + 1) * incx + 1 > size(x)) then
              write(*,'(a)') "Error: The vector is too small."
              stop
          else
              ix = (- n + 1) * incx + 1
          end if
      end if
      
      do i = 1, n
          if (x(ix) /= 0.0_dp) then
              absxi = abs(x(ix))
              if (scale < absxi) then
                  ssq = 1.0_dp + ssq * (scale / absxi)**2
                  scale = absxi
              else
                  ssq = ssq + (absxi / scale)**2
              end if
          end if
          ix = ix + incx
      end do
      dnrm2 = scale * sqrt(ssq)

      return
end function dnrm2

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! DSCAL scales a vector by a constant.
!
!  Discussion:
!    This routine uses real(8) arithmetic.
!
!  Modified:
!    08 April 1999
!
!  Author:
!    Jack Dongarra
!    Fortran90 translation by John Burkardt.
!
!  Reference:
!    Jack Dongarra, Jim Bunch, Cleve Moler, Pete Stewart,
!    LINPACK User's Guide,
!    SIAM, 1979,
!    ISBN13: 978-0-898711-72-1,
!    LC: QA214.L56.
!
!    Charles Lawson, Richard Hanson, David Kincaid, Fred Krogh,
!    Algorithm 539, 
!    Basic Linear Algebra Subprograms for Fortran Usage,
!    ACM Transactions on Mathematical Software,
!    Volume 5, Number 3, September 1979, pages 308-323.
!
!  Parameters:
!
!    Input, integer N, the number of entries in the vector.
!
!    Input, real ( kind = 8 ) SA, the multiplier.
!
!    Input/output, real ( kind = 8 ) X(*), the vector to be scaled.
!
!    Input, integer INCX, the increment between successive entries of X.
!

      subroutine  dscal(n,sa,x,incx)
      use lsqrDataModule
      implicit none
      integer,       intent(in)    :: n,incx
      real(kind=dp), intent(in)    :: sa
      real(kind=dp), intent(inout) :: x(:)

      integer :: i,ix

      if (n <= 0) then
         return
      end if

      if (0 <= incx) then ! For positive increments
          if (1 + (n - 1) * incx > size(x)) then
              write(*,'(a)') "Error: The vector is too small."
              stop
          else
              ix = 1
          end if
      else
          if ((-n + 1) * incx + 1 > size(x)) then
              write(*,'(a)') "Error: The vector is too small."
              stop
          else
              ix = (-n + 1) * incx + 1 ! For negative increments
          end if
      end if

      do i = 1, n
          x(ix) = sa * x(ix)
          ix = ix + incx
      end do

      return
end subroutine dscal
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
