!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!

program main

! Reads the species and config
use mod_io,       only : read_input
! Reads the grid
use mod_grid,     only : read_grid
! Set KIM potential
use mod_forces,   only : set_forces
! Do Projection
use mod_rigidity, only : set_rigidity
! Calculating the stress
use mod_stress,   only : set_stress, compute_stress
! Control output
use mod_output,   only : string, output, outputmode, logunit, logentry, logoutput, &
                         nerror, error, erroroutput
use mod_file,     only : GetFreeUnit
use mod_global,   only : cmdunit, dp
use mod_qclib,    only : downcase

implicit none

!-------------------------------------------------------------------------------
! local variables
!
character(len=512) :: filename, input
character(len=4)   :: command
logical            :: comments
integer            :: ierror, i, l
real(kind=dp)      :: program_start, program_end

call cpu_time(program_start)

!-------------------------------------------------------------------------------
! Print out the MDStressLab Information
! Open the log file
! Process the command the input
!
! set current output mode for subroutine output
outputmode = 'normal'

!-------------------------------------------------------------------------------
! setup banner
!
string = ' '; call output
string = '----------------------------------------' //  &
         '------------------------------------'
call output
string = ' '; call output
write(string,'(''M D S t r e s s L a b'')')
call output
string = ' '; call output
string = ' '; call output
string = 'This code is distributed under the ' //       &
         'CDDL open source license'
call output
string = ' '; call output
string = 'Authors: N. C. Admal, Min Shi and E. B. Tadmor ' //    &
         '(University of Minnesota)'
call output
string = ' '; call output
string = 'Version 1.2.0'
call output
string = ' '; call output
string = 'For more information contact tadmor@umn.edu'
call output
string = ' '; call output
string = '----------------------------------------' //  &
         '------------------------------------'
call output
string = ' '; call output

! Announce the start of simulation
filename='sc.log'
logunit=GetFreeUnit()
open(logunit,file=filename,status='replace',form='formatted')
call logentry
write(string,'(a)') 'Simulation started.'
call logoutput

!-------------------------------------------------------------------------------
! parse input file: strip leading spaces from lines, strip comment lines.
!
filename='sc.cmd'
cmdunit=GetFreeUnit()

open(cmdunit,file=filename,status='replace',form='formatted')
do
    read(5,'(a512)',iostat=ierror) input
    if (ierror .ne. 0) exit
    input = trim(adjustl(input))

    if (input(1:1) .eq. '%') cycle ! it is a comment
    if (input == '') cycle ! it is an empty line

    i = 1
    comments = .false.
    do while (i .lt. 512) ! Every line is supposed to be <= 512 characters
        i = i + 1
        if (input(i:i) .eq. '%') then
            comments = .true.
            exit
        end if
    enddo

    if (comments) then
        write(cmdunit,'(a)') trim(input(1:i - 1))
    else
        write(cmdunit,'(a)') trim(input(1:i))
    end if
enddo
rewind(cmdunit)

!-------------------------------------------------------------------------------
! read a card and compare first 4 columns with macro list
!
do
    read(cmdunit,'(a512)',iostat=ierror) input

    if (ierror .ne. 0) then
        nerror = 2  
        write(error(1),'(a)') 'Something wrong with the command input.'
        write(error(2),'(a)') 'Macro stop must be called to terminate the simulation.'
        call erroroutput
    end if

    command = input(1:4)

    select case(trim(downcase(command)))
    !-------------------------------------------------------------------------------
    ! macro 'read' : Read all the input files
    !
    case('read')
        call read_input
    !-------------------------------------------------------------------------------
    ! macro 'grid' : Set up the grid
    !
    case('grid')
        call read_grid
    !-------------------------------------------------------------------------------
    ! macro 'potential' : Interatomic force settings
    !
    case('pote')
        call set_forces
    !-------------------------------------------------------------------------------
    ! macro 'uniqueness' : Set up projection
    !
    case('uniq')
        call set_rigidity
    !-------------------------------------------------------------------------------
    ! macro 'stress' : Calculate stresses
    !
    case('stre')
        call set_stress
        call compute_stress
    !-------------------------------------------------------------------------------
    ! macro 'stop' :  Terminate the program
    !
    case('stop')
        exit
    case default
        nerror = 1 
        error(1) = 'stage "' // command // '" not recognized.**'
        call erroroutput
    end select

end do

! Announce program termination
call logentry
write(string,'(a)') 'Simulation ended.'
call logoutput

! Close all open files
do l = 7,99
    close(l)
enddo

! Delete the temporary file 'sc.cmd'
open(unit=7, iostat=ierror, file='sc.cmd', status='old')
if (ierror == 0) close(7, status='delete')

call Terminate_All

call cpu_time(program_end)

write(string,'(a,F25.16,a)') 'MDStressLab runs in ', program_end - program_start, ' s.'; call output

contains

    subroutine Terminate_All
    use mod_output, only : string, output, heading
    use mod_particles, only : deallocate_atoms
    use mod_box, only : deallocate_species
    use mod_grid, only : deallocate_grid
    use mod_neigh, only : deallocate_neigh
    use mod_rigidity, only : deallocate_rigidity
    use mod_forces, only : deallocate_forces
    use mod_stress, only : deallocate_stress
    implicit none

    heading=.true.
    write(string,'(a)')'**Deallocating data**'; call output

    call deallocate_atoms
    call deallocate_species
    call deallocate_grid
    call deallocate_neigh
    call deallocate_rigidity
    call deallocate_forces
    call deallocate_stress

    end subroutine Terminate_All

end program
