!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!e***************************************************************************
!**
!**   MODULE mod_box: ensemble properties of the simulation box
!**
!**   type definitions:
!**   ----------------
!**   type(species) :: spec(:)
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   samplesize(:)
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   nat      - number of atoms
!**   dimn     - dimension
!**   nsp      - number of species in the sample configuration
!**   nsp_file - number of species in species file
!**   maxnsp   - maximum number of species
!**  
!**   logical Variable Definitions:
!**   ----------------------------
!**   pbc - periodic boundary conditions
!**
!**   Contains Routines:
!**   ------------------
!**   allocate_box : allocates memory to the variable samplesize
!****************************************************************************
module mod_box
use mod_global, only : dp
use mod_types,  only : species
implicit none
save
private
public allocate_box,       &
       allocate_species,   &
       deallocate_species, &
       samplesize,         & ! samplesize is the default sample size
       samplesize_i,       & ! i -> initial
       samplesize_f,       & ! f -> final
       pbc,                &
       dimn,               &
       nat,                &
       spec,               &
       nsp,                &
       spec_file,          &
       nsp_file

real(kind=dp), dimension(3) :: samplesize,samplesize_i,samplesize_f
logical, dimension(3) :: pbc
integer :: dimn,nat,nsp,nsp_file
character(len=3), allocatable, dimension(:) :: spec
type(species), allocatable, target, dimension(:) :: spec_file

contains
    !-------------------------------------------------------------------------------
    ! allocate_species -
    !
    subroutine allocate_species(sfile)
    use mod_file, only : FileExists,iofile
    use mod_output, only : string,output,nerror,error,erroroutput,heading
    implicit none
    character(len=512),intent(in) :: sfile
    character(len=3) :: junk_id
    real(kind=dp) :: junk_mass
    integer :: i,ierror,iunit

    if (FileExists(sfile,.false.)) then
        call iofile(sfile,'formatted  ',iunit,.false.,.false.)
        heading=.true.
        write(string,'(a)') '**Reading species data**'
        call output
    else   
        nerror = 2
        error(1) = 'species information file not found'
        error(2) = 'in subroutine allocate_species, filename = ' &
                   // trim(sfile)
        call erroroutput
    endif

    nsp_file = 0
    do 
        read(iunit,*,iostat=ierror) junk_id,junk_mass
        if (ierror .ne. 0) then
            exit
        endif
        nsp_file=nsp_file+1
        if (nsp_file > 200) then
            nerror = 1
            write(error(1),'(3a)') &
            'Number of species in the file: ', trim(sfile), ' exceeds maximum tolerated = 200.'
            call erroroutput
        endif
    end do

    if (nsp_file == 0) then
        nerror = 1
        write(error(1),'(2a)') trim(sfile), ' does not contain any valid species!'
        call erroroutput
    end if

    allocate(spec(nsp_file))
    allocate(spec_file(nsp_file))

    rewind(iunit)

    write(string,'(a)') 'Warning: If same species names occur, only the first one will take effect in computation!'
    call output

    write(string,'(a)') 'Species on file:'; call output

    do i = 1, nsp_file
        read(iunit,*,iostat=ierror) spec_file(i)%element, spec_file(i)%mass
        if (ierror .ne. 0) then
            exit
        endif
        write(string,2010) spec_file(i)%element, spec_file(i)%mass; call output
    end do
    write(string,'(a,I5)') 'Total number of species in file = ',nsp_file; call output
    !-------------------------------------------------------------------------------
    ! Format descriptors
    !
    2010 format(2x,a3,2x,f25.16)
    end subroutine allocate_species

    !-------------------------------------------------------------------------------
    ! allocate_box -
    !
    subroutine allocate_box(cfile)
    use mod_file, only : FileExists,iofile
    use mod_output, only : string,output,nerror,error,erroroutput,heading
    implicit none
    character(len=512),intent(in) :: cfile
    integer :: i,ierror,iunit

    if (FileExists(cfile,.false.)) then
        call iofile(cfile,'formatted  ',iunit,.false.,.false.)
        heading=.true.
        write(string,'(a)') '**Reading configuration data**'
        call output
    else   
        nerror = 2
        error(1) = 'configuration information file not found'
        error(2) = 'in subroutine allocate_box, filename = ' &
                   // trim(cfile)
        call erroroutput
    endif

    read(iunit,*,iostat=ierror) dimn,nat

    if (ierror /= 0) then
        nerror = 1
        write(error(1),'(a)') 'Error reading dimn and nat.'
        call erroroutput
    end if

    if (nat <= 0) then
        nerror = 2
        error(1) = 'Error configuration in file: '
        write(error(2),'(a,I0,a)') 'Number of atoms = ', nat, ', which should be > 0.'
        call erroroutput
    endif

    if (dimn /= 3) then
        nerror = 4
        error(1) = 'Error configuration in file: '
        write(error(2),'(a,I0,a)') 'Dimension = ', dimn, '. This code could only work when dimn = 3.'
        write(error(3),'(a)') 'Even for 2D material, you need to specify a thickness.'
        write(error(4),'(a)') 'Otherwise, the neighbor list and KIM will not work.'
        call erroroutput
    endif

    write(string,2000) 'Dimension        = ',dimn; call output
    write(string,2000) 'Number of atoms  = ',nat; call output

    read(iunit,*, iostat=ierror) samplesize_i,samplesize_f

    if (ierror /= 0) then
        nerror = 1
        write(error(1),'(a)') 'Error reading sample size.'
        call erroroutput
    end if

    do i = 1, dimn
        if (samplesize_i(i) <= 0.0_dp) then
            nerror = 1
            write(error(1),'(a,I0,a,F25.16)') 'Initial sample size should be > 0. Your input of direction ',&
                                               i ,' is ', samplesize_i(i)
            call erroroutput
        end if
        if (samplesize_f(i) <= 0.0_dp) then
            nerror = 1
            write(error(1),'(a,I0,a,F25.16)') 'Final sample size should be > 0. Your input of direction ',&
                                               i ,' is ', samplesize_f(i)
            call erroroutput
        end if
    enddo

    ! The default sample size is set as final sample size
    samplesize = samplesize_f
    write(string,2020) samplesize_i; call output
    write(string,2030) samplesize; call output
    read(iunit,*,iostat=ierror) pbc
    if (ierror .ne. 0) then
        nerror = 1
        write(error(1),'(a)') 'Error reading periodic boundary conditions.'
        call erroroutput
    endif
    write(string,2040) pbc; call output
    !-------------------------------------------------------------------------------
    ! format descriptors
    !
    2000 format(a,I10)
    2020 format('Reference sample size     = ',3F25.16)
    2030 format('Deformed sample size      = ',3F25.16)
    2040 format('PBC                       = ',3L15)
    end subroutine allocate_box

    !-------------------------------------------------------------------------------
    ! deallocate_box
    !
    subroutine deallocate_species
    implicit none
    if (allocated(spec)) deallocate(spec)
    if (allocated(spec_file)) deallocate(spec_file)
    end subroutine deallocate_species

end module mod_box

