!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**  MODULE mod_file : contains routines relevant to file handling
!**
!**
!**  Variable Definitions:
!**  ---------------------
!**
!**  none
!**
!**
!**  Contains Routines:
!**  ------------------
!**
!**  FileExists   - check for the existence of a file
!**  GetFreeUnit  - get a unit number not being currently used
!**  iofile       - opens files and returns a unit number (uses getfreeunit)
!**
!****************************************************************************

module mod_file
implicit none
save
private
public FileExists, GetFreeUnit, iofile
contains

   !------------------------------------------------------------------
   ! FileExists -- returns .true. if file 'filename' exists in the current 
   !               directory
   !
   !      Passed Parameters :
   !          filename (in) : filename
   !             toprt (in) : logical variable, set to .true. if function
   !                          should print a message if file is not found
   !
   !      Module Parameters :
   !                      none
   !
   !      Algorithm :
   !            self explanatory
   !
   !      Notes :
   !          none
   !
   !      Author :
   !            E.B.Tadmor (12/31/97)
   !
   !      Revisions :
   !              none
   !
   !--
   logical function FileExists(filename,toprt)
      use mod_output, only : string, output
      implicit none

      !** Transferred Variables **|
      character(len=*), intent(in) :: filename
      logical, intent(in) :: toprt

      inquire(file=filename,exist=FileExists)
      if (.not.FileExists.and.toprt) then
         write(string,'(3a)') '* WARNING: File ',trim(filename),' not found.'
         call output
      endif
      return
   end function FileExists

   !------------------------------------------------------------------
   ! GetFreeUnit -- return the number of a unit number currently not
   !                being used.
   !
   !      Passed Parameters :
   !                      none
   !
   !      Module Parameters :
   !                      none
   !
   !      Algorithm :
   !            Loop over all possible unit numbers until an unallocated
   !            one is found, otherwise report an error and stop.
   !
   !      Notes :
   !          none
   !
   !      Author :
   !            E.B.Tadmor (12/31/97)
   !
   !      Revisions :
   !              none
   !
   !--
   integer function GetFreeUnit()

      use mod_output, only : nerror, error, erroroutput
      implicit none

      !** Local Variables **|
      logical InUse
!
! reserved file numbers: 5 = std in
!                        6 = std out
!                       99 = abort.dat
!
      do GetFreeUnit=7,98
         inquire(unit=GetFreeUnit,opened=InUse)
         if (.not. InUse) return
      enddo
      nerror = 1
      error(1) = 'Could not obtain a free unit handle in function GetFreeUnit!'
      call erroroutput
   end function GetFreeUnit

   !------------------------------------------------------------------
   ! iofile -- if filename is open, returns the unit number, otherwise
   !           gets a free unit number and opens filename.
   !
   !      Passed Parameters :
   !          filename (in) : filename
   !            format (in) : character string for file format
   !             toprt (in) : logical flag indicating whether to print or run
   !                          silently
   !            append (in) : logical flag indicating whether to open the file
   !                          in append mode
   !            logic (out) : unit number  
   !
   !      Module Parameters :
   !                      uses GetFreeUnit
   !
   !      Algorithm :
   !            self explanatory
   !
   !      Notes :
   !          none
   !
   !      Author :
   !            R.E.Miller (09/25/95)
   !
   !      Revisions :
   !            Added append flag (ET 01/23/05)
   !
   !--
      subroutine iofile(filename,format,logic,toprt,append)
      use mod_output, only : string, output
      implicit none
!
!---- io file management: open file and return unit number
!
      character(len=512) filename
      character(len=11) format
      integer logic
      logical ofl,toprt,append
!
      inquire(file=filename,opened=ofl)
      if (.not.ofl) then
         logic=GetFreeUnit()
         if (append) then
            open(logic,file=filename,status='unknown',form=format, &
                 position='append')
         else
            open(logic,file=filename,status='unknown',form=format)
         endif
         if (toprt) then
            write(string,'(4a)')'* Opening ',trim(format),' file: ', trim(filename)
            call output
         endif
      else
         inquire(file=filename,number=logic)
         if (toprt) then
            write(string,'(2a)')'* Accessing previously opened file: ', trim(filename)
            call output
         endif
      end if
      return
      end subroutine iofile

end module mod_file

