!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**   MODULE mod_forces: 
!**
!**   type definitions:
!**   ----------------
!**   N/A
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   fij(:)  - interatomic forces
!**   pfij(:) - interatomic forces
!**   scl
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   max_pairs_per_atom - max number of interactions per atom
!**   list_length        - number of interactions
!**   marker1,marker2    -
!**   list
!**  
!**   logical Variable Definitions:
!**   ----------------------------
!**
!**   Contains Routines:
!**   ------------------
!**   ngh_list(range) - Form the neighbor list, including all atom pairs 
!**                     that are within a distance "range" to each other.
!**                     "range", in our case is equal to the interaction 
!**                     range.
!**   compute_forces - 
!****************************************************************************
#include "KIM_API_status.h"
module mod_forces
use, intrinsic :: iso_c_binding
use KIM_API_F03
use mod_global, only : dp
implicit none
save
private
public compute_forces,          &
       set_forces,              &
       deallocate_forces,       &
       rcutoff

real(kind=dp)                            :: energy, rcutoff
character(len=KIM_KEY_STRING_LENGTH)     :: modelname
!#define KIM_KEY_STRING_LENGTH 128 (in KIM)
real(kind=dp),dimension(:,:),allocatable :: fi

contains

    subroutine set_forces
    !-------------------------------------------------------------------------------
    ! SUBROUTINE set_forces :
    !
    ! Passed Parameters :-
    ! -----------------
    !
    ! Non-Obvious Parameter :-
    ! ---------------------
    !
    use mod_global, only : cmdunit
    use mod_output, only : string,output,erroroutput,nerror,error,heading
    use mod_qclib,  only : next,downcase,freein
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=512) :: input
    character(len=4)  :: cc
    integer :: lower,upper,ierror
    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    heading=.true.
    write(string,'(a)')'**Reading interatomic potential**'; call output
    do
        read(cmdunit,'(a512)',iostat=ierror) input
        if (ierror .ne. 0) then
            nerror = 2
            write(error(1),'(a)') 'Something wrong with the command input.'
            write(error(2),'(a)') 'Macro end must be called to continue the simulation.'
            call erroroutput
        end if
        lower = 0
        upper = next(lower,input)
        cc = input(lower+1:upper-1)
        
        select case(trim(downcase(cc))) 
        case('modl')
            lower = upper
            upper = next(lower,input)
            modelname= trim(input(lower+1:upper-1))
            write(string,'(2a)') 'Model name is: ', modelname; call output
        case('end')
            exit
        case default
            nerror = 1
            error(1) = 'macro "' // cc // '" not recognized in stage potential**'
            call erroroutput
        end select
    enddo

    end subroutine set_forces

    subroutine compute_forces
    !-------------------------------------------------------------------------------
    ! SUBROUTINE compute_forces :
    !     Calculates forces on the atoms by invoking the kim_api
    !     Calculates interatomic forces through standard, projection
    !
    ! Passed Parameters :-
    ! -----------------
    !
    ! Non-Obvious Parameter :-
    ! ---------------------
    !
    use mod_particles, only : pos,element,particle_energy,particleTypes
    use mod_box,       only : nat,nsp,pbc,dimn,samplesize
    use mod_neigh,     only : form_at_neigh_list,neighObjFull,nbc,skin
    use mod_fij,       only : allocate_fij
    use mod_rigidity,  only : project,project_forces
    use mod_file,      only : FileExists,iofile
    use mod_output,    only : string,output
    use mod_kim
    use KIM_API_F03
    implicit none

    character(len=KIM_KEY_STRING_LENGTH) :: testkimfile="descriptor.kim"
    type(c_funptr)     :: pdedr,pget_neigh
    type(c_ptr)        :: pneighObjFull

    write(string,'(a)') 'Starting force calculation : '; call output
    allocate(fi(dimn,nat))
    !-------------------------------------------------------------------------------
    ! Setup the KIM API object
    ! 1. Broadcast the
    !     a. name of the test and model
    !     b. nature of neigh list
    !     c. dimension
    !     d. number of atoms
    !     e. position of atoms
    !     f. number of species
    !     g. samplesize.
    ! 2. Get the cutoff radius from the model
    !

    ! Calculate Forces and interatomic forces using model via KIM_API
    allocate(particle_energy(nat))
    allocate(particleTypes(nat))
    call setup_KIM_API_object(testkimfile, modelname, &
                              dimn,nat,nsp,pos,element,particle_energy,particleTypes,energy,fi,rcutoff,samplesize,nbc)
                          
    write(string,5009) rcutoff; call output

    ! Reset pbc to false if boundary conditions are not periodic
    if (nbc .ne. 0 .and. nbc .ne. 1) then
        pbc = .false.
        write(string,5010) pbc; call output
    endif

    call compatibility_check

    !-------------------------------------------------------------------------------
    ! Neighbor list calculation for the given cutoff
    !
    call form_at_neigh_list(skin*rcutoff)

    !-------------------------------------------------------------------------------
    ! Allocate interatomic forces fij
    !
    call allocate_fij

    !-------------------------------------------------------------------------------
    ! Inform KIM where it can find the neighbor list and the get_neigh function
    !
    pdedr= c_funloc(process_dedr)
    pget_neigh = c_funloc(get_neigh)
    pneighObjFull = c_loc(neighObjFull)
    call setup_kim_access(pdedr,pget_neigh,pneighObjFull)

    call model_compute

    !-------------------------------------------------------------------------------
    ! Free memory
    !
    call free_KIM_API_object

    !-------------------------------------------------------------------------------
    ! Compute proj_fij using rigidity theory
    !
    if (project) call project_forces

    5009 format('Cutoff radius from model = ',f5.3)
    5010 format('CAUTION: PBC not supported by the model. Resetting pbc = ',3L15)
    end subroutine compute_forces

!-------------------------------------------------------------------------------
!
! get_neigh neighbor list access function
!
! This function implements Locator and Iterator mode
!
!-------------------------------------------------------------------------------
function get_neigh(pkim,mode,request,part,numnei,pnei1part, &
                                  pRij) bind(c)
  use KIM_API_F03
  use mod_types, only : neighObjFullType
  implicit none

  !-- Transferred variables
  type(c_ptr),    intent(in)  :: pkim
  integer(c_int), intent(in)  :: mode
  integer(c_int), intent(in)  :: request
  integer(c_int), intent(out) :: part
  integer(c_int), intent(out) :: numnei
  type(c_ptr),    intent(out) :: pnei1part
  type(c_ptr),    intent(out) :: pRij
  integer(c_int)              :: get_neigh
  !-- Local variables
  integer(c_int), save :: iterVal = 0
  integer(c_int)  partToReturn
  integer(c_int), pointer         :: numberOfParticles; type(c_ptr) :: pnParts
  type(neighObjFullType), pointer :: neighObjFull;      type(c_ptr) :: pneighObjFull
  integer(c_int)  ier, idum

  ! unpack number of particles
  pnParts = kim_api_get_data(pkim, "numberOfParticles", ier)
  if (ier.lt.KIM_STATUS_OK) then
     idum = kim_api_report_error(__LINE__, __FILE__, "kim_api_get_data", ier)
     stop
  endif
  call c_f_pointer(pnParts, numberOfParticles)

  ! unpack neighbor list object
  pneighObjFull = kim_api_get_data(pkim, "neighObject", ier)
  if (ier.lt.KIM_STATUS_OK) then
     idum = kim_api_report_error(__LINE__, __FILE__, &
                                 "kim_api_get_data", ier)
     stop
  endif
  call c_f_pointer(pneighObjFull, neighObjFull)

  ! check mode and request
  if (mode.eq.0) then ! iterator mode
     if (request.eq.0) then ! reset iterator
        iterVal = 0
        get_neigh = KIM_STATUS_NEIGH_ITER_INIT_OK
        return
     elseif (request.eq.1) then ! increment iterator
        iterVal = iterVal + 1
        if (iterVal.gt.numberOfParticles) then
           get_neigh = KIM_STATUS_NEIGH_ITER_PAST_END
           return
        else
           partToReturn = iterVal
        endif
     else
        idum = kim_api_report_error(__LINE__, __FILE__, &
                                    "Invalid request in get_neigh", &
                                    KIM_STATUS_NEIGH_INVALID_REQUEST)
        get_neigh = KIM_STATUS_NEIGH_INVALID_REQUEST
        return
     endif
  elseif (mode.eq.1) then ! locator mode
     if ( (request.gt.numberOfParticles) .or. (request.lt.1)) then
        idum = kim_api_report_error(__LINE__, __FILE__, &
                                    "Invalid part ID in get_neigh", &
                                    KIM_STATUS_PARTICLE_INVALID_ID)
        get_neigh = KIM_STATUS_PARTICLE_INVALID_ID
        return
     else
        partToReturn = request
     endif
  else ! not iterator or locator mode
     idum = kim_api_report_error(__LINE__, __FILE__, &
                                 "Invalid mode in get_neigh", &
                                 KIM_STATUS_NEIGH_INVALID_MODE)
     get_neigh = KIM_STATUS_NEIGH_INVALID_MODE
     return
  endif

  ! set the returned part
  part = partToReturn

  ! set the returned number of neighbors for the returned part
  numnei = neighObjFull%neighborList(1,part)

  ! set the location for the returned neighbor list
  pnei1part = c_loc(neighObjFull%neighborList(2,part))

  pRij = c_null_ptr

  get_neigh = KIM_STATUS_OK
  return
end function get_neigh

    integer(c_int) function process_dedr(pkim,deidr,r,dx,at_i,at_j) bind(c)
    use KIM_API_F03
    use mod_fij,       only : process_dedr_fij
    implicit none

    ! Passed variables
    type(c_ptr),   intent(in) :: pkim,dx
    real(c_double),intent(in) :: deidr,r
    integer(c_int),intent(in) :: at_i,at_j

    if (at_i < at_j) then 
        call process_dedr_fij(deidr,at_i,at_j)
    else
        call process_dedr_fij(deidr,at_j,at_i)
    endif

    process_dedr = KIM_STATUS_OK
    end function process_dedr

    ! Check the model (The final sample size) must satisfy MIOPBC if PBC is T (samplesize_f > 2 * rcutoff)
    subroutine compatibility_check
    use mod_box, only : dimn,samplesize_f,pbc
    use mod_output, only : erroroutput,nerror,error
    implicit none

    ! Local variables
    integer :: i

    do i = 1, dimn
        if (samplesize_f(i) <= 2.0_dp * rcutoff .and. pbc(i)) then
            if (i == 1) then
                nerror = 1
                write(error(1),'(a,F25.16,a,F25.16,a)') 'In x direction, samplesize_f, ',samplesize_f(i), &
                                                        ' is <= 2 * rcutoff: ', 2.0_dp * rcutoff, &
                                                        ' ,violating MIOPBC.'
                call erroroutput
            else if (i == 2) then
                nerror = 1
                write(error(1),'(a,F25.16,a,F25.16,a)') 'In y direction, samplesize_f, ',samplesize_f(i), &
                                                        ' is <= 2 * rcutoff: ', 2.0_dp * rcutoff, &
                                                        ' ,violating MIOPBC.'
                call erroroutput
            else if (i == 3) then
                nerror = 1
                write(error(1),'(a,F25.16,a,F25.16,a)') 'In z direction, samplesize_f, ',samplesize_f(i), &
                                                        ' is <= 2 * rcutoff: ', 2.0_dp * rcutoff, &
                                                        ' ,violating MIOPBC.'
                call erroroutput
            else
                nerror = 1
                write(error(1),'(a)') 'Something funny happens. dimn /= 3.'
                call erroroutput
            end if
        end if
    end do

    end subroutine compatibility_check

    subroutine deallocate_forces
    use mod_fij, only : deallocate_fij
    implicit none
    if (allocated(fi)) deallocate(fi)
    call deallocate_fij
    end subroutine deallocate_forces

end module mod_forces
