!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
! mod_neigh builds the atom neighbor list and the grid neighbor list
!

module mod_neigh
use mod_global, only : dp
use mod_types,  only : neighObjFullType, bin_at, bin_grid, bin_neighbor_list
implicit none
save
private
public  max_pairs_per_atom,    &
        marker1,marker2,       &
        list,list_length,      &
        ngh_list,              &
        neighObjFull,          &
        full_grid_verlet_list, &
        form_at_neigh_list,    &
        form_grid_neigh_list,  &
        deallocate_neigh,      &
        nbc,                   &
        skin

! Neighbor list for calculation of forces
integer,dimension(:),allocatable                 :: marker1,marker2
integer,allocatable,target                       :: list(:)
integer                                          :: list_length
integer                                          :: max_pairs_per_atom
real(kind=dp),parameter                          :: skin = 1.2_dp
integer                                          :: nbc
logical,dimension(:),allocatable                 :: InList ! used to remove duplicates
type(neighObjFullType),target                    :: neighObjFull
type(bin_neighbor_list),dimension(:),allocatable :: full_grid_verlet_list
integer                                          :: debug = 0

contains

    !-------------------------------------------------------------------------------
    ! Construct neighbor lists for force calculation
    !
    ! Note: This is a copy of neighbor list construction implemented in QC3D
    !       using binning method by Min Shi in 2015, but does not include the optimal
    !       method which is subdividing the bins as it is disliked by Woo Kyun
    !       and maybe we do not need this here.
    !       The current method in QC3D uses hash table to save space considering
    !       the sparseness of local nodes.
    !       However, here this is purely atomistic. We could copy the hash table
    !       here in the future if we want.
    !       The reason why I did not copy the hash table method is because that one
    !       is still being discussed, tested and not well understood.
    !
    !       If some bugs occurs, pls use the old brute force method.
    !
    !       Need Nikhil Revision
    !
    ! TODO: THIS COULD DEFINITELY BE PARALLELIZED
    subroutine form_at_neigh_list(cutoff)
    use mod_box,       only : dimn, nat, samplesize, pbc
    use mod_particles, only : pos
    use mod_output,    only : string, output, nerror, error, erroroutput
    use mod_file,      only : GetFreeUnit
    use mod_qclib,     only : qsorti
    implicit none

    ! TODO: Replace the neighObjFull with verlet list to reduce memory
    ! Passed Variables
    real(kind=dp), intent(in)      :: cutoff

    ! Local Variables
    real(kind=dp), dimension(dimn) :: pos_max, pos_min, rij
    integer :: max_list_length, total_neigh, kount
    integer :: nxbin, nybin, nzbin, nbin
    integer :: nxatombin, nyatombin, nzatombin, natombin
    integer :: ibin
    integer :: jindex, kindex, lindex
    integer :: i, j, iatom, jatom
    real(kind=dp) :: dxbin, dybin, dzbin

    ! The formula to get the right bin is i +-2;1;0 +-2x;x;0 +-2xy;xy;0 +-xyz;0
    integer, dimension(5*5*5*3) :: neighbin
    integer :: unit
    integer, dimension(:), allocatable :: tmp_neighbor_list,junk

    type(bin_neighbor_list), dimension(:), allocatable :: full_verlet_list
    type(bin_neighbor_list), dimension(:), allocatable :: half_verlet_list
    type(bin_neighbor_list), dimension(:), allocatable :: tmp_verlet_list ! used to sort
    type(bin_neighbor_list), dimension(:), allocatable :: seq_verlet_list ! used to sort
    type(bin_at),            dimension(:), allocatable :: verlet_bin

    write(string,'(a)') 'Constructing atom neighbor list'; call output

    max_pairs_per_atom = 1000 ! First as a guess used in tmp_neigh_list, then after we find out the true value
                              ! We set it as what it should be.
    kount = 0

    if (allocated(full_verlet_list)) deallocate(full_verlet_list)
    if (.not. allocated(full_verlet_list)) allocate(full_verlet_list(nat))
    if (allocated(half_verlet_list)) deallocate(half_verlet_list)
    if (.not. allocated(half_verlet_list)) allocate(half_verlet_list(nat))
    if (allocated(tmp_verlet_list)) deallocate(tmp_verlet_list)
    if (.not. allocated(tmp_verlet_list)) allocate(tmp_verlet_list(nat))
    if (allocated(seq_verlet_list)) deallocate(seq_verlet_list)
    if (.not. allocated(seq_verlet_list)) allocate(seq_verlet_list(nat))
    if (allocated(InList)) deallocate(InList)
    if (.not. allocated(InList)) allocate(InList(nat))

    pos_max = -huge(1.0_dp)
    pos_min = huge(1.0_dp)

    ! Find the suitable place to create the bins
    do i = 1, nat
        where (pos(:,i) > pos_max)
            pos_max = pos(:,i)
        end where
        where (pos(:,i) < pos_min)
            pos_min = pos(:,i)
        end where
    end do

    pos_max = pos_max + 1.0e-10_dp
    pos_min = pos_min - 1.0e-10_dp

    write(string,'(a,3F25.16)') "Building bins in this region MAX: ", pos_max; call output
    write(string,'(a,3F25.16)') "Building bins in this region MIN: ", pos_min; call output

    ! PBC actually questionable as Woo Kyun said remove epsilon?
    nxbin = max(floor((pos_max(1) - pos_min(1)) / (epsilon(1.0_dp) + cutoff)),1)
    nybin = max(floor((pos_max(2) - pos_min(2)) / (epsilon(1.0_dp) + cutoff)),1)
    nzbin = max(floor((pos_max(3) - pos_min(3)) / (epsilon(1.0_dp) + cutoff)),1)
    dxbin = (pos_max(1) - pos_min(1)) / real(nxbin,dp)
    dybin = (pos_max(2) - pos_min(2)) / real(nybin,dp)
    dzbin = (pos_max(3) - pos_min(3)) / real(nzbin,dp)

    nbin = nxbin * nybin * nzbin

    write(string,'(a,I0,a,3F25.16)') "nxbin: ", nxbin, "; dxbin: ", dxbin; call output
    write(string,'(a,I0,a,3F25.16)') "nybin: ", nybin, "; dybin: ", dybin; call output
    write(string,'(a,I0,a,3F25.16)') "nzbin: ", nzbin, "; dzbin: ", dzbin; call output
    write(string,'(a,I0)') "nbin: ", nbin; call output
    allocate(verlet_bin(nbin))
    verlet_bin(:)%natoms = 0

    do i = 1, nat
        nxatombin = ceiling((pos(1,i) - pos_min(1)) / dxbin)
        nyatombin = ceiling((pos(2,i) - pos_min(2)) / dybin)
        nzatombin = ceiling((pos(3,i) - pos_min(3)) / dzbin)
        if (nxatombin <= 0 .or. nyatombin <= 0 .or. nzatombin <= 0) then
            nerror = 3
            error(1) = "Something wrong with the nxyzatombins! It is smaller than 0!"
            write(error(2),'(a,I0,a,I0,a,I0)') "nxatombin: ", nxatombin, " nyatombin: ",&
                                                nyatombin, " nzatombin: ", nzatombin
            write(error(3),'(a,3F25.16)') "Offending atoms: ", pos(:,i)
            call erroroutput
        endif
        if (nxatombin > nxbin .or. nyatombin > nybin .or. nzatombin > nzbin) then
            nerror = 3
            error(1) = "Something wrong with the nxyzatombins! It goes out of bounds!"
            write(error(2),'(a,I0,a,I0,a,I0)') "nxatombin: ", nxatombin, " nyatombin: ",&
                                                nyatombin, " nzatombin: ", nzatombin
            write(error(3),'(a,3F25.16)') "Offending atoms: ", pos(:,i)
            call erroroutput
        endif
        natombin = (nzatombin - 1) * nxbin * nybin + (nyatombin - 1) * nxbin + nxatombin
        verlet_bin(natombin)%natoms = verlet_bin(natombin)%natoms + 1
    enddo

    do ibin = 1, nbin
        allocate(verlet_bin(ibin)%bin_atomlist(verlet_bin(ibin)%natoms))
        verlet_bin(ibin)%natoms = 0
    end do

    do i = 1, nat
        nxatombin = ceiling((pos(1,i) - pos_min(1)) / dxbin)
        nyatombin = ceiling((pos(2,i) - pos_min(2)) / dybin)
        nzatombin = ceiling((pos(3,i) - pos_min(3)) / dzbin)
        natombin = (nzatombin - 1) * nxbin * nybin + (nyatombin - 1) * nxbin + nxatombin
        verlet_bin(natombin)%natoms = verlet_bin(natombin)%natoms + 1
        verlet_bin(natombin)%bin_atomlist(verlet_bin(natombin)%natoms) = i
    enddo

    allocate(tmp_neighbor_list(max_pairs_per_atom))

    full_verlet_list(:)%nneigh = 0
    InList = .false.

    do ibin = 1, nbin
        do iatom = 1, verlet_bin(ibin)%natoms
            neighbin = ibin
            neighbin(1) = neighbin(1) - 2
            neighbin(2) = neighbin(2) - 1
            neighbin(3) = neighbin(3)
            neighbin(4) = neighbin(4) + 1
            neighbin(5) = neighbin(5) + 2

            do jindex = 6,10
                neighbin(jindex) = neighbin(jindex-5) - 2 * nxbin
            enddo

            do jindex = 11,15
                neighbin(jindex) = neighbin(jindex-10) - 1 * nxbin
            enddo

            do jindex = 16,20
                neighbin(jindex) = neighbin(jindex-15) + 1 * nxbin
            enddo

            do jindex = 21,25
                neighbin(jindex) = neighbin(jindex-20) + 2 * nxbin
            enddo

            do kindex = 26,50
                neighbin(kindex) = neighbin(kindex-25) - 2 * nxbin * nybin
            enddo

            do kindex = 51,75
                neighbin(kindex) = neighbin(kindex-50) - 1 * nxbin * nybin
            enddo

            do kindex = 76,100
                neighbin(kindex) = neighbin(kindex-75) + 1 * nxbin * nybin
            enddo

            do kindex = 101,125
                neighbin(kindex) = neighbin(kindex-100) + 2 * nxbin * nybin
            enddo

            do lindex = 126,250
                neighbin(lindex) = neighbin(lindex-125) + nxbin * nybin * nzbin
            enddo
            do lindex = 251,375
                neighbin(lindex) = neighbin(lindex-250) - nxbin * nybin * nzbin
            enddo

            do i = 1,5*5*5*3
                if (neighbin(i) <= 0 .or. neighbin(i) > nbin ) then ! cycle the wrong neighbin id
                    cycle
                endif
                do jatom = 1, verlet_bin(neighbin(i))%natoms
                    if (ibin == neighbin(i) .and. iatom == jatom) cycle ! cycle the same atom
                    rij = pos(:,verlet_bin(neighbin(i))%bin_atomlist(jatom)) &
                        - pos(:,verlet_bin(ibin)%bin_atomlist(iatom))
                    where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                        rij = rij-sign(samplesize,rij)
                    end where
                    !if (sqrt(dot_product(rij,rij)) <= cutoff) then
                    if (dot_product(rij,rij) <= cutoff * cutoff) then

                        if (.not. InList(verlet_bin(neighbin(i))%bin_atomlist(jatom))) then

                            full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh = &
                            full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh + 1

                            if (full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh > max_pairs_per_atom) then
                                allocate(junk(max_pairs_per_atom))
                                junk = tmp_neighbor_list
                                deallocate(tmp_neighbor_list)
                                allocate(tmp_neighbor_list(max_pairs_per_atom * 2))
                                do j = 1, max_pairs_per_atom
                                    tmp_neighbor_list(j) = junk(j)
                                end do
                                max_pairs_per_atom =  max_pairs_per_atom * 2
                                write(string,'(a,I0)') "Performance alert: # of Neighbors increased to ", max_pairs_per_atom
                                call output
                                deallocate(junk)
                            end if

                            tmp_neighbor_list(full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh) = &
                            verlet_bin(neighbin(i))%bin_atomlist(jatom)
                            InList(verlet_bin(neighbin(i))%bin_atomlist(jatom)) = .true.
                        end if
                    endif
                enddo
            enddo

            allocate(full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))&
              %neigh(full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh))
            allocate(half_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))&
              %neigh(full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh))
            allocate(tmp_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))&
              %neigh(full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh))
            allocate(seq_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))&
              %neigh(full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh))

            do j = 1, full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%nneigh
                full_verlet_list(verlet_bin(ibin)%bin_atomlist(iatom))%neigh(j) = tmp_neighbor_list(j)
                InList(tmp_neighbor_list(j)) = .false.
            enddo
            deallocate(tmp_neighbor_list)
            allocate(tmp_neighbor_list(max_pairs_per_atom))
        enddo
    enddo

    if (allocated(tmp_neighbor_list)) deallocate(tmp_neighbor_list)
    if (allocated(verlet_bin)) deallocate(verlet_bin)
    if (allocated(InList)) deallocate(InList)

    max_pairs_per_atom = -1

    do i = 1, nat
        if (max_pairs_per_atom < full_verlet_list(i)%nneigh) then
            max_pairs_per_atom = full_verlet_list(i)%nneigh
        end if

        call qsorti(full_verlet_list(i)%neigh,seq_verlet_list(i)%neigh,full_verlet_list(i)%nneigh)

        do j = 1, full_verlet_list(i)%nneigh
            tmp_verlet_list(i)%neigh(j) = full_verlet_list(i)%neigh(seq_verlet_list(i)%neigh(j))
        enddo

        do j = 1, full_verlet_list(i)%nneigh
            full_verlet_list(i)%neigh(j) = tmp_verlet_list(i)%neigh(j)
        enddo
    enddo

    !max_pairs_per_atom = maxval(full_verlet_list(:)%nneigh)

    write(string,'(a,I0)') "max_pairs_per_atom: ", max_pairs_per_atom; call output

    if (allocated(tmp_verlet_list)) deallocate(tmp_verlet_list)
    if (allocated(seq_verlet_list)) deallocate(seq_verlet_list)

    half_verlet_list(:)%nneigh = 0
    do i = 1, nat
        do j = 1, full_verlet_list(i)%nneigh
            if (full_verlet_list(i)%neigh(j) > i) then
                half_verlet_list(i)%nneigh = half_verlet_list(i)%nneigh + 1
                half_verlet_list(i)%neigh(half_verlet_list(i)%nneigh) = full_verlet_list(i)%neigh(j)
            end if
        end do
    end do

    max_list_length = max_pairs_per_atom * nat
    if (.not. (allocated(marker1))) allocate(marker1(nat))
    if (.not. (allocated(marker2))) allocate(marker2(nat))
    if (.not. (allocated(list))) allocate(list(max_list_length))
    if (.not. (allocated(neighObjFull%neighborList))) allocate(neighObjFull%neighborList(max_pairs_per_atom+1,nat))

    marker1 = 0
    marker2 = 0
    list = 0
    neighObjFull%neighborList = 0
    total_neigh = 0
    marker1(1) = 1

    do i = 1,nat - 1
        total_neigh = total_neigh + half_verlet_list(i)%nneigh
        marker1(i + 1) = total_neigh + 1
        marker2(i) = total_neigh
    enddo

    total_neigh = total_neigh + half_verlet_list(nat)%nneigh
    marker2(nat) = total_neigh

    kount = 0
    do i = 1, nat
        do j = 1, half_verlet_list(i)%nneigh
            kount = kount + 1
            list(kount) = half_verlet_list(i)%neigh(j)
        enddo
    enddo

    if (kount /= total_neigh) then
        nerror = 1
        write(error(1),'(a)') "Something went wrong when counting the neighbor"
        call erroroutput
    endif

    if (nbc .eq. 0 .or. nbc .eq. 2) then   ! MI_OPBC_H or NEIGH_PURE_H Half_neighborlist
        do i = 1, nat
            neighObjFull%neighborList(1,i) = half_verlet_list(i)%nneigh
            do j = 1, half_verlet_list(i)%nneigh
                neighObjFull%neighborList(1 + j,i) = half_verlet_list(i)%neigh(j)
            enddo
        enddo
    else ! full neighborlist
        do i = 1, nat
            neighObjFull%neighborList(1,i) = full_verlet_list(i)%nneigh
            do j = 1, full_verlet_list(i)%nneigh
                neighObjFull%neighborList(1 + j,i) = full_verlet_list(i)%neigh(j)
            enddo
        enddo
    endif

    list_length = total_neigh

    write(string,'(a,I0)') "Number of atoms: ", nat; call output

    ! Debug stuff
    if (debug /= 0) then
        list_length = 0
        unit = GetFreeUnit()
        open(unit,file='ngh_list_bin.txt',status='replace',form='formatted')

        do i = 1, nat
            write(unit,*) "marker1: ", i
            write(unit,*) marker1(i)
        enddo

        do i = 1, nat
            write(unit,*) "marker2: ", i
            write(unit,*) marker2(i)
        enddo

        do i = 1, nat
            write(unit,*) "List: ", i
            write(unit,*) list(marker1(i):marker2(i))
        enddo

        call flush(unit)
        close(unit)

        unit = GetFreeUnit()
        open(unit,file='full_ngh_list_bin.txt',status='replace',form='formatted')

        do i = 1, nat
            write(unit,*) "Number of neighbors: ", i
            write(unit,*) neighObjFull%neighborList(1,i)
        enddo

        do i = 1, nat
            write(unit,*) "neighbors: ", i
            write(unit,*) neighObjFull%neighborList(2:max_pairs_per_atom+1,i)
        enddo
        call flush(unit)
        close(unit)

        call deallocate_neigh

        max_list_length = max_pairs_per_atom * nat
        if (.not. (allocated(marker1))) allocate(marker1(nat))
        if (.not. (allocated(marker2))) allocate(marker2(nat))
        if (.not. (allocated(list))) allocate(list(max_list_length))
        if (.not. (allocated(neighObjFull%neighborList))) allocate(neighObjFull%neighborList(max_pairs_per_atom+1,nat))
        marker1 = 0
        marker2 = 0
        list = 0
        neighObjFull%neighborList = 0
        call ngh_list(list,list_length,marker1,marker2,cutoff)
        call full_ngh_list(cutoff)
    endif

    if (allocated(full_verlet_list)) deallocate(full_verlet_list)
    if (allocated(half_verlet_list)) deallocate(half_verlet_list)

    write(string,'(a)') "Neighbor List Construction Finish!"; call output
    write(string,5010) list_length; call output
    5010 format('Allocating vector fij of size list_length, which is the total # of neighbors = ',I10)

    end subroutine form_at_neigh_list

    !-------------------------------------------------------------------------------
    ! Construct neighbor list for grid point to speed up the stress calculation
    !
    ! Note: This is a copy of neighbor list construction implemented in QC3D
    !       using binning method by Min Shi in 2015, but does not include the optimal
    !       method which is subdividing the bins as it is disliked by Woo Kyun
    !       and maybe we do not need this here.
    !       The current method in QC3D uses hash table to save space considering
    !       the sparseness of local nodes.
    !       However, here this is purely atomistic. We could copy the hash table
    !       here in the future if we want.
    !       The reason why I did not copy the hash table method is because that one
    !       is still being discussed, tested and not well understood.
    !
    !       If some bugs occurs, pls use the old brute force method.
    !
    !       Need Nikhil Revision
    !
    ! TODO: THIS COULD DEFINITELY BE PARALLELIZED
    subroutine form_grid_neigh_list(rsum)
    use mod_box, only : dimn,nat,samplesize,pbc
    use mod_particles, only : pos
    use mod_grid, only : gpt,ngrid
    use mod_output, only : string, output, nerror, error, erroroutput
    use mod_file, only : GetFreeUnit
    use mod_qclib, only : qsorti
    implicit none

    ! Passed Variables
    real(kind=dp), dimension(dimn), intent(in) :: rsum

    ! Local Variables
    real(kind=dp), dimension(dimn) :: pos_max, pos_min, rij
    integer :: i, j, igrid, jatom
    integer :: max_atoms_per_grid
    integer :: nxbin, nybin, nzbin, nbin
    real(kind=dp) :: dxbin, dybin, dzbin
    integer :: nxatombin, nyatombin, nzatombin, natombin
    integer :: nxgridbin, nygridbin, nzgridbin, ngridbin
    integer :: ibin
    integer :: jindex, kindex, lindex
    ! The formula to get the right bin is i +-2;1;0 +-2x;x;0 +-2xy;xy;0 +-xyz;0
    integer, dimension(5*5*5*3) :: neighbin

    integer, dimension(:), allocatable :: tmp_grid_neighbor_list,junk
    type(bin_neighbor_list), dimension(:), allocatable :: tmp_grid_verlet_list ! used to sort
    type(bin_neighbor_list), dimension(:), allocatable :: seq_grid_verlet_list ! used to sort
    type(bin_grid),          dimension(:), allocatable :: verlet_bin_grid

    write(string,'(a)') 'Constructing grid neighbor list'; call output

    max_atoms_per_grid = 20000 ! First as a guess used in tmp_grid_neigh_list, then after we find out the true value
                               ! We set it as what it should be.

    if (allocated(full_grid_verlet_list)) deallocate(full_grid_verlet_list)
    if (.not. allocated(full_grid_verlet_list)) allocate(full_grid_verlet_list(ngrid))
    if (allocated(tmp_grid_verlet_list)) deallocate(tmp_grid_verlet_list)
    if (.not. allocated(tmp_grid_verlet_list)) allocate(tmp_grid_verlet_list(ngrid))
    if (allocated(seq_grid_verlet_list)) deallocate(seq_grid_verlet_list)
    if (.not. allocated(seq_grid_verlet_list)) allocate(seq_grid_verlet_list(ngrid))
    if (allocated(InList)) deallocate(InList)
    if (.not. allocated(InList)) allocate(InList(nat))

    pos_max = -huge(1.0_dp)
    pos_min = huge(1.0_dp)

    ! Find the suitable place to create the bins
    ! GFILE MAY CAUSE USER-INPUT GRID POS

    do i = 1, nat
        where (pos(:,i) > pos_max)
            pos_max = pos(:,i)
        end where
        where (pos(:,i) < pos_min)
            pos_min = pos(:,i)
        end where
    end do

    do i = 1, ngrid
        where (gpt(:,i) > pos_max)
            pos_max = gpt(:,i)
        end where
        where (gpt(:,i) < pos_min)
            pos_min = gpt(:,i)
        end where
    end do

    pos_max = pos_max + 1e-10_dp
    pos_min = pos_min - 1e-10_dp

    write(string,'(a,3F25.16)') "Building bins in this region MAX: ", pos_max; call output
    write(string,'(a,3F25.16)') "Building bins in this region MIN: ", pos_min; call output

    ! PBC actually questionable as Woo Kyun said remove epsilon?
    nxbin = max(floor((pos_max(1) - pos_min(1)) / (epsilon(1.0_dp) + rsum(1))),1)
    nybin = max(floor((pos_max(2) - pos_min(2)) / (epsilon(1.0_dp) + rsum(2))),1)
    nzbin = max(floor((pos_max(3) - pos_min(3)) / (epsilon(1.0_dp) + rsum(3))),1)
    dxbin = (pos_max(1) - pos_min(1)) / real(nxbin,dp)
    dybin = (pos_max(2) - pos_min(2)) / real(nybin,dp)
    dzbin = (pos_max(3) - pos_min(3)) / real(nzbin,dp)

    nbin = nxbin * nybin * nzbin

    write(string,'(a,I0,a,3F25.16)') "nxbin: ", nxbin, "; dxbin: ", dxbin; call output
    write(string,'(a,I0,a,3F25.16)') "nybin: ", nybin, "; dybin: ", dybin; call output
    write(string,'(a,I0,a,3F25.16)') "nzbin: ", nzbin, "; dzbin: ", dzbin; call output
    write(string,'(a,I0)') "nbin: ", nbin; call output
    allocate(verlet_bin_grid(nbin))
    verlet_bin_grid(:)%natoms = 0
    verlet_bin_grid(:)%ngrid = 0

    do i = 1, ngrid
        nxgridbin = ceiling((gpt(1,i) - pos_min(1)) / dxbin)
        nygridbin = ceiling((gpt(2,i) - pos_min(2)) / dybin)
        nzgridbin = ceiling((gpt(3,i) - pos_min(3)) / dzbin)
        if (nxgridbin <= 0 .or. nygridbin <= 0 .or. nzgridbin <= 0) then
            nerror = 3
            error(1) = "Something wrong with the nxyzgridbins! It is smaller than 0!"
            write(error(2),'(a,I0,a,I0,a,I0)') "nxgridbin: ", nxgridbin, " nygridbin: ",&
                                                nygridbin, " nzgridbin: ", nzgridbin
            write(error(3),'(a,3F25.16)') "Offending grid: ", gpt(:,i)
            call erroroutput
        endif
        if (nxgridbin > nxbin .or. nygridbin > nybin .or. nzgridbin > nzbin) then
            nerror = 3
            error(1) = "Something wrong with the nxyzgridbins! It goes out of bounds!"
            write(error(2),'(a,I0,a,I0,a,I0)') "nxgridbin: ", nxgridbin, " nygridbin: ",&
                                                nygridbin, " nzgridbin: ", nzgridbin
            write(error(3),'(a,3F25.16)') "Offending grid: ", gpt(:,i)
            call erroroutput
        endif
        ngridbin = (nzgridbin - 1) * nxbin * nybin + (nygridbin - 1) * nxbin + nxgridbin
        verlet_bin_grid(ngridbin)%ngrid = verlet_bin_grid(ngridbin)%ngrid + 1
    enddo

    do ibin = 1, nbin
        allocate(verlet_bin_grid(ibin)%bin_gridlist(verlet_bin_grid(ibin)%ngrid))
        verlet_bin_grid(ibin)%ngrid = 0
    end do

    do i = 1, ngrid
        nxgridbin = ceiling((gpt(1,i) - pos_min(1)) / dxbin)
        nygridbin = ceiling((gpt(2,i) - pos_min(2)) / dybin)
        nzgridbin = ceiling((gpt(3,i) - pos_min(3)) / dzbin)
        ngridbin = (nzgridbin - 1) * nxbin * nybin + (nygridbin - 1) * nxbin + nxgridbin
        verlet_bin_grid(ngridbin)%ngrid = verlet_bin_grid(ngridbin)%ngrid + 1
        verlet_bin_grid(ngridbin)%bin_gridlist(verlet_bin_grid(ngridbin)%ngrid) = i
    enddo

    do i = 1, nat
        nxatombin = ceiling((pos(1,i) - pos_min(1)) / dxbin)
        nyatombin = ceiling((pos(2,i) - pos_min(2)) / dybin)
        nzatombin = ceiling((pos(3,i) - pos_min(3)) / dzbin)
        if (nxatombin <= 0 .or. nyatombin <= 0 .or. nzatombin <= 0) then
            nerror = 3
            error(1) = "Something wrong with the nxyzatombins! It is smaller than 0!"
            write(error(2),'(a,I0,a,I0,a,I0)') "nxatombin: ", nxatombin, " nyatombin: ",&
                                                nyatombin, " nzatombin: ", nzatombin
            write(error(3),'(a,3F25.16)') "Offending atoms: ", pos(:,i)
            call erroroutput
        endif
        if (nxatombin > nxbin .or. nyatombin > nybin .or. nzatombin > nzbin) then
            nerror = 3
            error(1) = "Something wrong with the nxyzatombins! It goes out of bounds!"
            write(error(2),'(a,I0,a,I0,a,I0)') "nxatombin: ", nxatombin, " nyatombin: ",&
                                                nyatombin, " nzatombin: ", nzatombin
            write(error(3),'(a,3F25.16)') "Offending atoms: ", pos(:,i)
            call erroroutput
        endif
        natombin = (nzatombin - 1) * nxbin * nybin + (nyatombin - 1) * nxbin + nxatombin
        verlet_bin_grid(natombin)%natoms = verlet_bin_grid(natombin)%natoms + 1
    enddo

    do ibin = 1, nbin
        allocate(verlet_bin_grid(ibin)%bin_atomlist(verlet_bin_grid(ibin)%natoms))
        verlet_bin_grid(ibin)%natoms = 0
    end do

    do i = 1, nat
        nxatombin = ceiling((pos(1,i) - pos_min(1)) / dxbin)
        nyatombin = ceiling((pos(2,i) - pos_min(2)) / dybin)
        nzatombin = ceiling((pos(3,i) - pos_min(3)) / dzbin)
        natombin = (nzatombin - 1) * nxbin * nybin + (nyatombin - 1) * nxbin + nxatombin
        verlet_bin_grid(natombin)%natoms = verlet_bin_grid(natombin)%natoms + 1
        verlet_bin_grid(natombin)%bin_atomlist(verlet_bin_grid(natombin)%natoms) = i
    enddo

    allocate(tmp_grid_neighbor_list(max_atoms_per_grid))

    full_grid_verlet_list(:)%nneigh = 0
    InList = .false.

    do ibin = 1, nbin
        do igrid = 1, verlet_bin_grid(ibin)%ngrid
            neighbin = ibin
            neighbin(1) = neighbin(1) - 2
            neighbin(2) = neighbin(2) - 1
            neighbin(3) = neighbin(3)
            neighbin(4) = neighbin(4) + 1
            neighbin(5) = neighbin(5) + 2

            do jindex = 6,10
                neighbin(jindex) = neighbin(jindex-5) - 2 * nxbin
            enddo

            do jindex = 11,15
                neighbin(jindex) = neighbin(jindex-10) - 1 * nxbin
            enddo

            do jindex = 16,20
                neighbin(jindex) = neighbin(jindex-15) + 1 * nxbin
            enddo

            do jindex = 21,25
                neighbin(jindex) = neighbin(jindex-20) + 2 * nxbin
            enddo

            do kindex = 26,50
                neighbin(kindex) = neighbin(kindex-25) - 2 * nxbin * nybin
            enddo

            do kindex = 51,75
                neighbin(kindex) = neighbin(kindex-50) - 1 * nxbin * nybin
            enddo

            do kindex = 76,100
                neighbin(kindex) = neighbin(kindex-75) + 1 * nxbin * nybin
            enddo

            do kindex = 101,125
                neighbin(kindex) = neighbin(kindex-100) + 2 * nxbin * nybin
            enddo

            do lindex = 126,250
                neighbin(lindex) = neighbin(lindex-125) + nxbin * nybin * nzbin
            enddo

            do lindex = 251,375
                neighbin(lindex) = neighbin(lindex-250) - nxbin * nybin * nzbin
            enddo

            do i = 1,5*5*5*3
                if (neighbin(i) <= 0 .or. neighbin(i) > nbin ) then ! cycle the wrong neighbin id
                    cycle
                endif
                do jatom = 1, verlet_bin_grid(neighbin(i))%natoms
                    rij = pos(:,verlet_bin_grid(neighbin(i))%bin_atomlist(jatom)) &
                        - gpt(:,verlet_bin_grid(ibin)%bin_gridlist(igrid))
                    where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                        rij = rij-sign(samplesize,rij)
                    end where
                    if (abs(rij(1)) <= rsum(1) .and. &
                        abs(rij(2)) <= rsum(2) .and. &
                        abs(rij(3)) <= rsum(3)) then
                        if (.not. InList(verlet_bin_grid(neighbin(i))%bin_atomlist(jatom))) then
                            full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%nneigh = &
                            full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%nneigh + 1

                            if (full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%nneigh &
                              > max_atoms_per_grid) then
                                allocate(junk(max_atoms_per_grid))
                                junk = tmp_grid_neighbor_list
                                deallocate(tmp_grid_neighbor_list)
                                allocate(tmp_grid_neighbor_list(max_atoms_per_grid * 2))
                                do j = 1, max_atoms_per_grid
                                    tmp_grid_neighbor_list(j) = junk(j)
                                end do
                                max_atoms_per_grid =  max_atoms_per_grid * 2
                                write(string,'(a,I0)') &
                                "Performance alert: Maximum # of Neighbors per grid increased to ", max_atoms_per_grid
                                call output
                                deallocate(junk)
                            end if

                            tmp_grid_neighbor_list(full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%nneigh) = &
                            verlet_bin_grid(neighbin(i))%bin_atomlist(jatom)
                            InList(verlet_bin_grid(neighbin(i))%bin_atomlist(jatom)) = .true.
                        end if
                    endif
                enddo
            enddo

            allocate(full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))&
              %neigh(full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%nneigh))
            allocate(tmp_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))&
              %neigh(full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%nneigh))
            allocate(seq_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))&
              %neigh(full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%nneigh))

            do j = 1, full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%nneigh
                full_grid_verlet_list(verlet_bin_grid(ibin)%bin_gridlist(igrid))%neigh(j) = tmp_grid_neighbor_list(j)
                InList(tmp_grid_neighbor_list(j)) = .false.
            enddo
            deallocate(tmp_grid_neighbor_list)
            allocate(tmp_grid_neighbor_list(max_atoms_per_grid))
        enddo
    enddo

    if (allocated(tmp_grid_neighbor_list)) deallocate(tmp_grid_neighbor_list)
    if (allocated(verlet_bin_grid)) deallocate(verlet_bin_grid)
    if (allocated(InList)) deallocate(InList)

    max_atoms_per_grid = -1

    do i = 1, ngrid
        if (max_atoms_per_grid < full_grid_verlet_list(i)%nneigh) then
            max_atoms_per_grid = full_grid_verlet_list(i)%nneigh
        end if

        call qsorti(full_grid_verlet_list(i)%neigh,seq_grid_verlet_list(i)%neigh,full_grid_verlet_list(i)%nneigh)
        do j = 1, full_grid_verlet_list(i)%nneigh
            tmp_grid_verlet_list(i)%neigh(j) = full_grid_verlet_list(i)%neigh(seq_grid_verlet_list(i)%neigh(j))
        enddo
        do j = 1, full_grid_verlet_list(i)%nneigh
            full_grid_verlet_list(i)%neigh(j) = tmp_grid_verlet_list(i)%neigh(j)
        enddo
    enddo

    if (allocated(tmp_grid_verlet_list)) deallocate(tmp_grid_verlet_list)
    if (allocated(seq_grid_verlet_list)) deallocate(seq_grid_verlet_list)

    write(string,'(a,I0)') "Max_atoms_per_grid: ", max_atoms_per_grid; call output

    write(string,'(a)') "Grid Neighbor List Construction Finish!"; call output

    end subroutine form_grid_neigh_list

    ! These are serial
    ! Old brute force half MDStressLab neighbor list, only used to debug now
    subroutine ngh_list(lst,lst_lth,mrk1,mrk2,cutoff)
    !-------------------------------------------------------------------------------
    ! SUBROUTINE ngh_list :
    !
    ! Passed Parameters :-
    ! -----------------
    ! lst (out)
    ! lst_lth (out)
    ! mrk1,mrk2 (out)
    !
    ! Non-Obvious Parameter :-
    ! ---------------------
    !
    use mod_particles, only : pos
    use mod_box, only : samplesize,pbc,nat,dimn
    use mod_output, only : string,output,erroroutput,nerror,error
    use mod_file, only : GetFreeUnit
    implicit none
    !-------------------------------------------------------------------------------
    ! Transferred Variables
    !
    integer, dimension(:), intent(inout) :: mrk1,mrk2,lst
    integer, intent(inout) :: lst_lth
    real(kind=dp), intent(in) :: cutoff
    !-------------------------------------------------------------------------------
    ! Local Variables
    !
    integer, dimension(:), allocatable :: advance
    real(kind=dp), dimension(dimn) :: rij
    real(kind=dp) :: rsqij,rangesq
    integer :: i,j,L
    integer :: unit

    real(kind=dp) checkpoint_neigh_brute_half_start,checkpoint_neigh_brute_half_end

    call cpu_time(checkpoint_neigh_brute_half_start)
    allocate(advance(nat))
    rangesq = cutoff**2
    L = 1
    do i = 1,nat
        do j = i+1,nat
            rij = pos(:,i) - pos(:,j)
            where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                rij = rij-sign(samplesize,rij)
            end where
            rsqij = dot_product(rij,rij)
            if (rsqij < rangesq) then
                advance(j) = 1
            else
                advance(j) = 0
            endif
        enddo
        mrk1(i) = L
        do j = i+1,nat
            if ( L > size(lst)) then
                nerror = 4
                write(error(1),'(a)') "Neighbor List: List too small. Increase"
                write(error(2),'(a)') "the parameter max_pairs_per_atom in"
                write(error(3),'(a)') "module mod_neigh. Stopped at atom "
                write(error(4),'(2I10)') i,mrk2(1)-mrk1(1)
                call erroroutput
            endif
            lst(L) = j
            L = L + advance(j)
        enddo
        mrk2(i) = L - 1
    enddo
    lst_lth = L - 1
    deallocate(advance)

    call cpu_time(checkpoint_neigh_brute_half_end)
    write(string,'(a,F25.16,a)') "Building neighbor list brute_half method cost time: ",&
    checkpoint_neigh_brute_half_end - checkpoint_neigh_brute_half_start,' s.'; call output
    unit = GetFreeUnit()
    open(unit,file='ngh_list.txt',status='replace',form='formatted')

    do i = 1, nat
        write(unit,*) "marker1: ", i
        write(unit,*) marker1(i)
    enddo

    do i = 1, nat
        write(unit,*) "marker2: ", i
        write(unit,*) marker2(i)
    enddo

    do i = 1, nat
        write(unit,*) "List: ", i
        write(unit,*) list(marker1(i):marker2(i))
    enddo

    call flush(unit)
    close(unit)

    end subroutine ngh_list

    ! Old brute force half/full KIM neighbor list, only used to debug now
    subroutine full_ngh_list(rcut)
    use mod_particles, only : pos
    use mod_box, only : samplesize,pbc,nat
    use mod_output, only : string,output,erroroutput,nerror,error
    use mod_file, only : GetFreeUnit
    implicit none

    !-- Transferred variables
    real(kind=dp),         intent(in)    :: rcut

    !-- Local variables
    integer ::  i, j, a
    real(kind=dp) dx(3)
    real(kind=dp) r2
    real(kind=dp) rcut2
    integer :: unit

    real(kind=dp) checkpoint_neigh_brute_full_start,checkpoint_neigh_brute_full_end

    call cpu_time(checkpoint_neigh_brute_full_start)

    rcut2 = rcut**2

    do i=1,nat
        a = 1
        do j=1,nat
            dx(:) = pos(:,j) - pos(:,i)
            where (abs(dx) > 0.5_dp*samplesize .and. pbc)  ! apply PBC
                dx = dx - sign(samplesize,dx)
            endwhere
            r2 = dot_product(dx, dx)
            if (r2.le.rcut2) then
               ! part j is a neighbor of part i
               if (nbc .eq. 0 .or. nbc .eq. 2) then   ! MI_OPBC_H or NEIGH_PURE_H
                   if ( i .lt. j ) then
                      a = a+1
                      neighObjFull%neighborList(a,i) = j
                   endif
               else
                   if (i.ne.j) then
                      a = a+1
                      neighObjFull%neighborList(a,i) = j
                   endif
               endif
            endif
        enddo

        if (a - 1 > max_pairs_per_atom) then
            nerror = 5
            write(error(1),'(a)') "Neighbor List: List too small. Increase"
            write(error(2),'(a)') "the parameter max_pairs_per_atom in"
            write(error(3),'(a)') "module mod_neigh. Stopped at atom "
            write(error(4),'(I0)') i
            write(error(5),'(a,I0,a)') "which has ", a - 1, " neighbors."
            call erroroutput
        endif

        ! part i has a-1 neighbors
        neighObjFull%neighborList(1,i) = a-1
    enddo

    call cpu_time(checkpoint_neigh_brute_full_end)
    write(string,'(a,F25.16,a)') "Building neighbor list brute_full method cost time: ",&
    checkpoint_neigh_brute_full_end - checkpoint_neigh_brute_full_start,' s.'; call output

    unit = GetFreeUnit()
    open(unit,file='full_ngh_list.txt',status='replace',form='formatted')

    do i = 1, nat
        write(unit,*) "Number of neighbors: ", i
        write(unit,*) neighObjFull%neighborList(1,i)
    enddo

    do i = 1, nat
        write(unit,*) "neighbors: ", i
        write(unit,*) neighObjFull%neighborList(2:max_pairs_per_atom+1,i)
    enddo

    return

    end subroutine full_ngh_list

    subroutine deallocate_neigh
    implicit none
    if (allocated(marker1)) deallocate(marker1)
    if (allocated(marker2)) deallocate(marker2)
    if (allocated(list)) deallocate(list)
    if (allocated(neighObjFull%neighborList)) deallocate(neighObjFull%neighborList)
    if (allocated(full_grid_verlet_list)) deallocate(full_grid_verlet_list)
    end subroutine deallocate_neigh

end module mod_neigh
