!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**   MODULE mod_particles: Particle properties
!**
!**   type definitions:
!**   ----------------
!**   N/A
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   pos(:), vel(:), pos_i(:), vel_i(:), pos_f(:), vel_f(:)
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   N/A
!**  
!**   logical Variable Definitions:
!**   ----------------------------
!**   N/A
!**
!**   Contains Routines:
!**   ------------------
!**   allocate_atom    : Allocates memory to the variable pos vel id
!**   read_config      : Reads in the positions, velocities and species of atoms
!**   deallocate_atoms : Deallocates atoms
!****************************************************************************
module mod_particles
use, intrinsic :: iso_c_binding
use mod_global, only : dp
implicit none
save
private
public allocate_atom,    &
       read_config,      &
       deallocate_atoms, &
       pos,pos_i,pos_f,vel,vel_i,vel_f,element,mass,particle_energy,particleTypes

real(kind=dp),    allocatable, dimension(:,:) :: pos,pos_i,pos_f,vel,vel_i,vel_f
real(kind=dp),    allocatable, dimension(:)   :: mass
character(len=3), allocatable, dimension(:)   :: element
real(kind=dp),    allocatable                 :: particle_energy(:)
integer(c_int),   allocatable                 :: particleTypes(:)

contains
!---------------------------------------------------------------------------
!   allocate_atom : allocates memory to atoms 
!
!   Passed Parameters :-
!              dimn (in) : dimension
!              nat  (in) : number of atoms
!
!   This is done in OpenMP style
!   Should be done in MPI style in the future
    subroutine allocate_atom(dimn,nat)
    implicit none
    integer, intent(in) :: dimn,nat

    allocate(pos(dimn,nat))
    allocate(vel(dimn,nat))
    allocate(element(nat))
    allocate(mass(nat))

    allocate(pos_i(dimn,nat))
    allocate(vel_i(dimn,nat))

    allocate(pos_f(dimn,nat))
    allocate(vel_f(dimn,nat))

    end subroutine allocate_atom
!---------------------------------------------------------------------------
!   read_config : Reads in the positions, velocities and species 
!
!   Non-Obvious Parameters :-
!
    subroutine read_config(cfile)
    use mod_box, only : dimn,nat,nsp,spec,nsp_file,spec_file
    use mod_file, only : FileExists,iofile
    use mod_output, only : string,output,nerror,error,erroroutput
    implicit none
    character(len=512), intent(in) :: cfile

    character(len=160) fomt
    character(len=3) :: element_check
    integer :: ierror,i=0,iunit,j,k
    logical :: species_match = .false.,species_already_exists = .false.

    call allocate_atom(dimn,nat)

    if (FileExists(cfile,.false.)) then
        call iofile(cfile,'formatted  ',iunit,.false.,.false.)
    else   
        nerror = 2
        error(1) = 'Configuration information file not found'
        error(2) = 'in subroutine read_config, filename = '& 
                    // trim(cfile)
        call erroroutput
    endif
        
    nsp=0
    do 
        i = i + 1
        if (i .le. nat) then
            read(iunit,*,iostat=ierror) element(i),pos_i(:,i),vel_i(:,i)
            if (ierror .ne. 0) then
                nerror=1
                write(error(1),'(a)') 'Premature end-of-file, Filename = '&
                                      // trim(cfile)
                call erroroutput
            endif

            ! Match the atomic species with the species id's in the species file
            species_match = .false.
            do j=1,nsp_file
                if (spec_file(j)%element .eq. element(i)) then
                    mass(i) = spec_file(j)%mass
                    species_match = .true.

                    ! See if the current species is different from the previous atomic species
                    ! If 'yes', then nsp = nsp+1
                    if (nsp .eq. 0) then
                        nsp=nsp+1
                        spec(nsp) = spec_file(j)%element
                    else
                        species_already_exists=.false.
                        do k=1,nsp
                            if (spec_file(j)%element .eq. spec(k)) then
                                species_already_exists = .true.
                                exit
                            endif
                        enddo
                        if (species_already_exists .eqv. .false.) then
                            nsp = nsp+1
                            spec(nsp) = spec_file(j)%element
                        endif
                    endif

                endif
            end do
            if (species_match .eqv. .false.) then
                nerror = 1
                write(error(1),1003) i,element
                call erroroutput
            endif

        elseif (i .gt. nat .and. i .le. 2*nat) then
            read(iunit,*,iostat=ierror) element_check,pos_f(:,i-nat),vel_f(:,i-nat)

            if (element_check /= element(i-nat)) then
                nerror=5
                write(error(1),'(a)') 'The species of final configuration does not match with initial configuration.'
                write(error(2),'(a)') 'Something is wrong with your input configuraion.'
                write(error(3),'(a,i0)') 'The index of the error atom is ', i-nat
                write(error(4),'(2a)') 'The initial species of this atom was: ', trim(adjustl(element(i-nat)))
                write(error(5),'(3a)') 'It changes to ', trim(adjustl(element_check)), ' in the final configuration.'
                call erroroutput
            endif

            if (ierror .ne. 0) then
                nerror=1
                write(error(1),'(a)') 'Premature end-of-file, Filename = '&
                                      // trim(cfile)
                call erroroutput
            endif
        elseif (i .gt. 2*nat) then
            read(iunit,*,iostat=ierror)
            if (ierror .eq. 0) then
                nerror=3
                write(error(1),'(a)') 'More atoms than specified!'
                write(error(2),1000) i-nat
                write(error(3),1001) nat
                call erroroutput
            else
                exit
            endif
        endif
    end do

    ! Check samplesize must be bigger than the real configuration
    call compatibility_check

    ! The current version of MPI is not memory scalable
    ! All the processors has a copy of atoms
    ! TODO: THIS COULD BE PARALLELIZED
    do i = 1, nat
        ! The default atomic positions and velocities are taken to be final coordinates and velocities
        ! This is changed to initial coordinates for calculating PK stress in mod_stress.f90
        pos(:,i) = pos_f(:,i)
        vel(:,i) = vel_f(:,i)
    enddo

    write(string,1004) nsp; call output
    write(string,'(a)') 'Species list in configuration: '; call output
    write(fomt,'(a,I3,a)') '(',nsp,'a)'
    write(string,fomt) spec(1:nsp); call output

    1000 format('i   = ',I10)
    1001 format('nat = ',I10)
    1003 format('Species of atom', I10,' = ',A3,' does not have a match in the species file')
    1004 format('Total number of species in the configuration file = ',I5)
    end subroutine read_config

    ! Check samplesize (the box) must be bigger than the real configuration
    subroutine compatibility_check
    use mod_box, only : nat,dimn,pbc,samplesize_i,samplesize_f
    use mod_output, only : nerror,error,erroroutput
    implicit none
    ! Local Variables
    real(kind=dp),dimension(dimn) :: pos_max_i, pos_min_i,& ! NOTE: THIS IS DONE AGAIN IN MOD_NEIGH.
                                     pos_max_f, pos_min_f ! WE SHOULD OPTIMIZE IT AND COULD DO IT PARALLEL
    integer :: i

    pos_max_i = -huge(1.0_dp)
    pos_min_i = huge(1.0_dp)
    pos_max_f = -huge(1.0_dp)
    pos_min_f = huge(1.0_dp)

    do i = 1, nat
        where (pos_i(:,i) > pos_max_i)
            pos_max_i = pos_i(:,i)
        end where
        where (pos_i(:,i) < pos_min_i)
            pos_min_i = pos_i(:,i)
        end where

        where (pos_f(:,i) > pos_max_f)
            pos_max_f = pos_f(:,i)
        end where
        where (pos_f(:,i) < pos_min_f)
            pos_min_f = pos_f(:,i)
        end where
    end do

    do i = 1, dimn
        if ((pos_max_i(i) - pos_min_i(i)) > samplesize_i(i) .and. pbc(i)) then
            if (i == 1) then
                nerror = 2
                write(error(1),'(a,F25.16,a,F25.16)') &
                                        'In x direction which is periodic, samplesize_i: ', samplesize_i(i), &
                                        ' is smaller than pos_max_i - pos_min_i : ', pos_max_i(i) - pos_min_i(i)
                write(error(2),'(a)')   ' ,which means some atoms are outside the box in the reference configuration.'
                call erroroutput
            else if (i == 2) then
                nerror = 2
                write(error(1),'(a,F25.16,a,F25.16)') &
                                        'In y direction which is periodic, samplesize_i: ', samplesize_i(i), &
                                        ' is smaller than pos_max_i - pos_min_i: ', pos_max_i(i) - pos_min_i(i)
                write(error(2),'(a)')   ' ,which means some atoms are outside the box in the reference configuration.'
                call erroroutput
            else if (i == 3) then
                nerror = 2
                write(error(1),'(a,F25.16,a,F25.16)') &
                                        'In z direction which is periodic, samplesize_i: ', samplesize_i(i), &
                                        ' is smaller than pos_max_i - pos_min_i: ', pos_max_i(i) - pos_min_i(i)
                write(error(2),'(a)')   ' ,which means some atoms are outside the box in the reference configuration.'
                call erroroutput
            else
                nerror = 1
                write(error(1),'(a)') 'Something funny happens. dimn /= 3.'
                call erroroutput
            end if
        end if
    enddo

    do i = 1, dimn
        if ((pos_max_f(i) - pos_min_f(i)) > samplesize_f(i) .and. pbc(i)) then
            if (i == 1) then
                nerror = 2
                write(error(1),'(a,F25.16,a,F25.16)') &
                                        'In x direction which is periodic, samplesize_f: ', samplesize_f(i), &
                                        ' is smaller than pos_max_f - pos_min_f: ', pos_max_f(i) - pos_min_f(i)
                write(error(2),'(a)')   ' ,which means some atoms are outside the box in the deformed configuration.'
                call erroroutput
            else if (i == 2) then
                nerror = 2
                write(error(1),'(a,F25.16,a,F25.16)') &
                                        'In y direction which is periodic, samplesize_f: ', samplesize_f(i), &
                                        ' is smaller than pos_max_f - pos_min_f: ', pos_max_f(i) - pos_min_f(i)
                write(error(2),'(a)')   ' ,which means some atoms are outside the box in the deformed configuration.'
                call erroroutput
            else if (i == 3) then
                nerror = 2
                write(error(1),'(a,F25.16,a,F25.16)') &
                                        'In z direction which is periodic, samplesize_f: ', samplesize_f(i), &
                                        ' is smaller than pos_max_f - pos_min_f: ', pos_max_f(i) - pos_min_f(i)
                write(error(2),'(a)')   ' ,which means some atoms are outside the box in the deformed configuration.'
                call erroroutput
            else
                nerror = 1
                write(error(1),'(a)') 'Something funny happens. dimn /= 3.'
                call erroroutput
            end if
        end if
    enddo

    end subroutine compatibility_check
!---------------------------------------------------------------------------
!   deallocate_atoms :  Deallocates atoms
!
!   Non-Obvious Parameters :-
!
    subroutine deallocate_atoms
    implicit none
    if (allocated(pos))     deallocate(pos)
    if (allocated(vel))     deallocate(vel)
    if (allocated(element)) deallocate(element)
    if (allocated(mass))    deallocate(mass)

    if (allocated(pos_i))   deallocate(pos_i)
    if (allocated(vel_i))   deallocate(vel_i)

    if (allocated(pos_f))   deallocate(pos_f)
    if (allocated(vel_f))   deallocate(vel_f)

    if (allocated(particleTypes)) deallocate(particleTypes)
    if (allocated(particle_energy)) deallocate(particle_energy)
    end subroutine deallocate_atoms

end module mod_particles
