!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!-------------------------------------------------------------------------------
!** QCLIB:  These are a lot of handy utilities and other routines that
!**   don't fit neatly elsewhere in the code.  Most are self
!**   explanatory.  Here is a brief table of contents, but note that the
!**   actual order of the routines is different.
!**
!** vector and matrix manipulations:
!**    cross_product   : 3-D cross product
!**    invmat33        : inverse of a 3x3 matrix
!**    detmat33        : determinant of a 3x3 matrix
!**    qsorti          : sort a list
!** character string manipulations:
!**    freein          : parse free-form input lines
!**    downcase        : make a string all lower case
!**    next            : find the next delimiter (comma or space) in a string
!**                      MDStressLab specific, but called in many places
!**
!************************************************************************

module mod_qclib
use mod_global, only : dp
implicit none
public

contains

  !************************************************************************
  !     Determinant of a 3x3 matrix
  !
  !--
  real(kind=dp) function detmat33_dp(a)

    implicit none

    !-- Transferred variables
    real(kind=dp), intent(in) :: a(3,3)

    detmat33_dp = a(1,1)*a(2,2)*a(3,3) + a(2,1)*a(3,2)*a(1,3) &
                + a(1,2)*a(2,3)*a(3,1) - a(3,1)*a(2,2)*a(1,3) &
                - a(1,2)*a(2,1)*a(3,3) - a(3,2)*a(2,3)*a(1,1)

    return

  end function detmat33_dp

  !************************************************************************
  !     Inverse of a 3x3 matrix
  !
  !--
  subroutine invmat33_dp(a,b)
    use mod_output, only : nerror,error,erroroutput
    implicit none

    !-- Transferred variables
    real(kind=dp), intent(in)  :: a(3,3)
    real(kind=dp), intent(out) :: b(3,3)

    !-- Local variables
    real(kind=dp) det

    det = detmat33_dp(a)
    if (abs(det) < epsilon(1.0_dp)) then
        nerror = 1
        write(error(1),'(a)') 'The determinant of the matrix is 0. Thus, cannot be inverted.'
        call erroroutput
    else
        b(1,1) = (a(2,2) * a(3,3) - a(2,3) * a(3,2)) / det
        b(1,2) = (a(1,3) * a(3,2) - a(1,2) * a(3,3)) / det
        b(1,3) = (a(1,2) * a(2,3) - a(2,2) * a(1,3)) / det
        b(2,1) = (a(2,3) * a(3,1) - a(2,1) * a(3,3)) / det
        b(2,2) = (a(1,1) * a(3,3) - a(1,3) * a(3,1)) / det
        b(2,3) = (a(2,1) * a(1,3) - a(1,1) * a(2,3)) / det
        b(3,1) = (a(2,1) * a(3,2) - a(2,2) * a(3,1)) / det
        b(3,2) = (a(3,1) * a(1,2) - a(1,1) * a(3,2)) / det
        b(3,3) = (a(1,1) * a(2,2) - a(1,2) * a(2,1)) / det
    end if

    return

  end subroutine invmat33_dp

  !************************************************************************
  !     Maxtrix 3 * 3 Multiply Vector 3 * 1
  !
  !--
  subroutine mat33vec31multiply_dp(a,b,c)
    implicit none

    !-- Transferred variables
    real(kind=dp), intent(in)   :: a(3,3)
    real(kind=dp), intent(in)   :: b(3)
    real(kind=dp), intent(out)  :: c(3)

    !-- Local variables

    c(1) = a(1,1) * b(1) + a(1,2) * b(2) + a(1,3) * b(3)
    c(2) = a(2,1) * b(1) + a(2,2) * b(2) + a(2,3) * b(3)
    c(3) = a(3,1) * b(1) + a(3,2) * b(2) + a(3,3) * b(3)

    return

  end subroutine mat33vec31multiply_dp

  !************************************************************************
  !     Maxtrix 3 * 3 Multiply Matrix 3 * 3
  !
  !--
  subroutine mat33mat33multiply_dp(a,b,c)
    implicit none

    !-- Transferred variables
    real(kind=dp), intent(in)   :: a(3,3)
    real(kind=dp), intent(in)   :: b(3,3)
    real(kind=dp), intent(out)  :: c(3,3)

    !-- Local variables
    integer :: i, j

    do i = 1, 3
        do j = 1, 3 
            c(i,j) = a(i,1) * b(1,j) + a(i,2) * b(2,j) + a(i,3) * b(3,j)
        end do
    end do
    return

  end subroutine mat33mat33multiply_dp

  ! NEED DOUBLE VERIFICATION THAT THIS ONE IS WORKING CORRECTLY WITH STEVE
  !*********************************************************************************
  !     The relation of a point and a Parallelepiped
  !
  !     Return code: 1 -> left
  !                  2 -> left boundary
  !                  3 -> in
  !                  4 -> right boundary
  !                  5 -> right
  !
  !     Input: a -> center
  !            b -> orientation
  !            c -> point
  !
  !--
  subroutine Point_Parallelepiped_Relation(a,b,c,d)

    implicit none
    !-- Transferred variables
    real(kind=dp), intent(in)  :: a(3)
    real(kind=dp), intent(in)  :: b(3,3)
    real(kind=dp), intent(in)  :: c(3)
    integer,       intent(out) :: d(3)

    !-- Local variables
    real(kind=dp) :: inv_b(3,3)
    real(kind=dp) :: equation(3)
    real(kind=dp) :: eps(3)
    integer       :: i

    eps(1) = 1.e-8_dp / (abs(b(1,1)) + abs(b(2,1)) + abs(b(3,1)))
    eps(2) = 1.e-8_dp / (abs(b(1,2)) + abs(b(2,2)) + abs(b(3,2)))
    eps(3) = 1.e-8_dp / (abs(b(1,3)) + abs(b(2,3)) + abs(b(3,3)))

    call invmat33_dp(b,inv_b)
    call mat33vec31multiply_dp(inv_b,c-a,equation)

    do i = 1, 3
        if (equation(i) < -1.0_dp - eps(i)) then
            d(i) = 1
        else if (equation(i) >= -1.0_dp - eps(i) .and. equation(i) <= -1.0_dp + eps(i)) then
            d(i) = 2
        else if (equation(i) > -1.0_dp + eps(i) .and. equation(i) < 1.0_dp - eps(i)) then
            d(i) = 3
        else if (equation(i) >= 1.0_dp - eps(i) .and. equation(i) <= 1.0_dp + eps(i)) then
            d(i) = 4
        else if (equation(i) > 1.0_dp + eps(i)) then
            d(i) = 5
        end if
    end do

    return

  end subroutine Point_Parallelepiped_Relation

  ! NEED DOUBLE VERIFICATION THAT THIS ONE IS WORKING CORRECTLY WITH STEVE
  !*********************************************************************************
  !     Determine whether a point is inside the double layered parallelepiped (including boundary)
  !
  !     Return code: 0 -> completely outside the parallelepiped (out of range)
  !                  1 -> completely inside the parallelepiped (zero equal within range)
  !                  2 -> on the surface of the parallelepiped (one equal)
  !                  4 -> on the edge of the parallelepiped (two equals)
  !                  8 -> on the vertex of the parallelepiped (three equals)
  !
  !     Input: a -> center
  !            b -> orientation
  !            c -> point
  !--
  integer function pointOnDoubleParallelepiped(a,b,c)

    implicit none
    !-- Transferred variables
    real(kind=dp), intent(in)  :: a(3)
    real(kind=dp), intent(in)  :: b(3,3)
    real(kind=dp), intent(in)  :: c(3)

    !-- Local variables
    real(kind=dp) :: inv_b(3,3)
    real(kind=dp) :: equation(3)
    real(kind=dp) :: eps(3)

    call invmat33_dp(b,inv_b)
    call mat33vec31multiply_dp(inv_b,c-a,equation)

    eps(1) = 1.e-8_dp / (abs(b(1,1)) + abs(b(2,1)) + abs(b(3,1)))
    eps(2) = 1.e-8_dp / (abs(b(1,2)) + abs(b(2,2)) + abs(b(3,2)))
    eps(3) = 1.e-8_dp / (abs(b(1,3)) + abs(b(2,3)) + abs(b(3,3)))

    if (any(abs(equation) > 1.0_dp + eps)) then
        pointOnDoubleParallelepiped = 0
    else
        pointOnDoubleParallelepiped = 1
        if (abs(abs(equation(1)) - 1.0_dp) <= eps(1)) then
            pointOnDoubleParallelepiped = pointOnDoubleParallelepiped * 2
        end if
        if (abs(abs(equation(2)) - 1.0_dp) <= eps(2)) then
            pointOnDoubleParallelepiped = pointOnDoubleParallelepiped * 2
        end if
        if (abs(abs(equation(3)) - 1.0_dp) <= eps(3)) then
            pointOnDoubleParallelepiped = pointOnDoubleParallelepiped * 2
        end if
    end if

    return

  end function pointOnDoubleParallelepiped

  ! NEED DOUBLE VERIFICATION THAT THIS ONE IS WORKING CORRECTLY WITH STEVE
  !*********************************************************************************
  !     Determine whether a point is inside the parallelepiped (including boundary)
  !
  !     Return code: 0 -> completely outside the parallelepiped (out of range)
  !                  1 -> completely inside the parallelepiped (zero equal within range)
  !                  2 -> on the surface of the parallelepiped (one equal)
  !                  4 -> on the edge of the parallelepiped (two equals)
  !                  8 -> on the vertex of the parallelepiped (three equals)
  !
  !     Input: a -> center
  !            b -> orientation
  !            c -> point
  !--
  integer function pointOnParallelepiped(a,b,c)

    implicit none
    !-- Transferred variables
    real(kind=dp), intent(in)  :: a(3)
    real(kind=dp), intent(in)  :: b(3,3)
    real(kind=dp), intent(in)  :: c(3)

    !-- Local variables
    real(kind=dp) :: inv_b(3,3)
    real(kind=dp) :: equation(3)

    call invmat33_dp(b,inv_b)
    call mat33vec31multiply_dp(inv_b,c-a,equation)

    if (any(abs(equation) > 1.0_dp)) then
        pointOnParallelepiped = 0
    else
        pointOnParallelepiped = 1
        if (abs(abs(equation(1)) - 1.0_dp) < epsilon(1.0_dp)) then
            pointOnParallelepiped = pointOnParallelepiped * 2
        end if
        if (abs(abs(equation(2)) - 1.0_dp) < epsilon(1.0_dp)) then
            pointOnParallelepiped = pointOnParallelepiped * 2
        end if
        if (abs(abs(equation(3)) - 1.0_dp) < epsilon(1.0_dp)) then
            pointOnParallelepiped = pointOnParallelepiped * 2
        end if
    end if

    return

  end function pointOnParallelepiped

! This one only works for 3D does not need to parallelized
subroutine cross_product(x1,x2,x3)
implicit none

!-- Transferred variables
real(kind=dp), intent(in)  :: x1(3), x2(3)
real(kind=dp), intent(out) :: x3(3)

x3(1) = x1(2) * x2(3) - x1(3) * x2(2)
x3(2) = x1(3) * x2(1) - x1(1) * x2(3)
x3(3) = x1(1) * x2(2) - x1(2) * x2(1)
return
end subroutine cross_product

! This one is only used in 3D does not need to parallelized
subroutine tensor_prod(dimn,vec1,vec2,tensor)
implicit none
integer, intent(in) :: dimn
real(kind=dp),intent(in) :: vec1(dimn),vec2(dimn)
real(kind=dp),intent(out):: tensor(dimn,dimn)
integer :: i,j
do i=1,dimn
    do j=1,dimn
        tensor(j,i) = vec1(j)*vec2(i)
    enddo
enddo
end subroutine tensor_prod

!** CHARACTER STRING MANIPULATIONS **!

!************************************************************************
integer function next(last,input)
implicit none
 
!-- Transferred variables
character(len=512), intent(in) :: input
integer, intent(in)            :: last
 
!-- Local Variables
integer i

do i = last + 1,512
   if ((input(i:i).eq.',') .or. (input(i:i).eq.' ')) then
!   if (input(i:i).eq.',') then
      next = i
      return
   end if
enddo
next = 513

return
end function next

!************************************************************************   
function downcase(s1)
implicit none

!-- Transferred variables
character(len=*), intent(in)  :: s1
character(len=len(s1)) downcase

!-- Local variables
integer i,ic

do i=1,len(s1)
   ic=ichar(s1(i:i))
   if ((ic > 64).and.(ic < 91)) then
      downcase(i:i)=achar(ic+32)
   else
      downcase(i:i)=s1(i:i)
   endif
enddo
end function downcase
!************************************************************************   
subroutine freein(input,ll,lr,intvar,realvar,isw)
implicit none

!-- Transferred variables
integer ll,lr,intvar,isw
real(kind=dp) realvar
character(len=512) input
 
!-- Local Variables
character(len=1) range1
character(len=2) range2
character(len=512) strinp
integer ldif
 
strinp = input((ll+1):(lr-1))
ldif = lr - ll - 1
if (ldif.eq.0) then
   if (isw.eq.1) intvar = 0
   if (isw.eq.2) realvar = 0.0_dp
else
   if (ldif.lt.10) then
      write(range1,'(i1)') ldif
      if (isw.eq.1) read(strinp,'(i'//range1//')') intvar
      if (isw.eq.2) read(strinp,'(f'//range1//'.0)') realvar
   else
      write(range2,'(i2)') ldif
      if (isw.eq.1) read(strinp,'(i'//range2//')') intvar
      if (isw.eq.2) read(strinp,'(f'//range2//'.0)') realvar
   end if
end if
return
end subroutine freein

    !-------------------------------------------------------------------------------
    ! color outputs colored charaters for emphasize.
    !
    !  Discussion:
    !
    !    This function outputs colored charaters for emphasize.
    !
    !  Modified:
    !
    !    08/11/2016
    !
    !  Author:
    !
    !    Min Shi
    !
    !  Parameters: 
    !  
    !  str = your character to output 
    !
    !  code:   black = '30'
    !          red = '31'
    !          green = '32'
    !          yellow = '33'
    !          blue = '34'
    !          magenta = '35'
    !          cyan = '36'
    !          white = '37'
    !          etc.
    !
    function color(str, code) result(out)
       implicit none
       character(len=*), intent(in) :: str
       character(len=*), intent(in) :: code
       character(len=:), allocatable :: out
       out = achar(27) // '[' // code // 'm' // str // achar(27) // '[0m'
    end function color

    ! This subroutine is not used anywhere.
    ! Assuming checking some intersection in QC3D
    subroutine intersect(arr1,arr2,arr12)
    implicit none

    integer, dimension(:),intent(in) :: arr1,arr2
    integer, dimension(:),intent(inout) :: arr12

    integer, dimension(size(arr1)) :: larr1
    integer, dimension(size(arr2)) :: larr2
    integer i,j,k,n

    larr1=arr1
    larr2=arr2

    call QsortC(larr1)
    call QsortC(larr2)

    k=1
    n=0
    arr12 = 0
    do i =1,size(larr1)
        do j=k,size(larr2)
            if (larr2(j) .le. larr1(i)) then
                if (larr2(j) .eq. larr1(i)) then
                    n=n+1
                    arr12(n) = larr1(i)
                    k=j+1
                    exit
                endif
            else
                exit
            endif
        enddo
    enddo
    end subroutine intersect

    ! Looks like a quick sort algorithm used in intersect
    ! However not used anywhere except in intersect
    recursive subroutine QsortC(A)
    implicit none
    integer, intent(in out), dimension(:) :: A
    integer :: iq

    if(size(A) > 1) then
        call Partition(A, iq)
        call QsortC(A(:iq-1))
        call QsortC(A(iq:))
    endif
    end subroutine QsortC

    ! Looks like used in quick sort
    subroutine Partition(A, marker)
    implicit none
    integer, intent(in out), dimension(:) :: A
    integer, intent(out) :: marker
    integer :: i, j
    integer:: temp
    integer:: x      ! pivot point
    x = A(1)
    i= 0
    j= size(A) + 1

    do
        j = j-1
        do
            if (A(j) <= x) exit
            j = j-1
        end do
        i = i+1
        do
            if (A(i) >= x) exit
            i = i+1
        end do
        if (i < j) then
            ! exchange A(i) and A(j)
            temp = A(i)
            A(i) = A(j)
            A(j) = temp
        elseif (i == j) then
            marker = i+1
            return
        else
            marker = i
            return
        endif
    end do
    end subroutine Partition

    ! This need to be parallelized
    subroutine amux ( n, x, y, a, ja, ia )
    !-------------------------------------------------------------------------------
    ! AMUX multiplies a CSR matrix A times a vector.
    !
    !  Discussion:
    !
    !    This routine multiplies a matrix by a vector using the dot product form.
    !    Matrix A is stored in compressed sparse row storage.
    !
    !  Modified:
    !
    !    07 January 2004
    !
    !  Author:
    !
    !    Youcef Saad
    !
    !  Parameters:
    !
    !    Input, integer N, the row dimension of the matrix.
    !
    !    Input, real X(*), and array of length equal to the column dimension
    !    of A.
    !
    !    Input, real A(*), integer JA(*), IA(NROW+1), the matrix in CSR
    !    Compressed Sparse Row format.
    !
    !    Output, real Y(N), the product A * X.
    !
    implicit none

    integer n

    real(kind = dp) a(*)
    integer i
    integer ia(*)
    integer ja(*)
    integer k
    real(kind = dp) t
    real(kind = dp) x(*)
    real(kind = dp) y(n)

    do i = 1, n
        !
        !  Compute the inner product of row I with vector X.
        !
        !t = 0.0D+00
        t = 0.0_dp
        do k = ia(i), ia(i + 1) - 1
            t = t + a(k) * x(ja(k))
        end do

        y(i) = t

    end do

    return
    end subroutine amux

    ! This need to be parallelized
subroutine atmux ( m ,n, x, y, a, ja, ia )
!-------------------------------------------------------------------------------
!  ATMUX computes A' * x for a CSR matrix A.
!
!  Discussion:
!
!    This routine multiplies the transpose of a matrix by a vector when the
!    original matrix is stored in compressed sparse row storage. Can also be
!    viewed as the product of a matrix by a vector when the original
!    matrix is stored in the compressed sparse column format.
!
!  Modified:
!
!    07 January 2004
!
!  Author:
!
!    Youcef Saad
!
!  Parameters:
!
!    Input, integer N, the row dimension of the matrix A'.
!
!    Input, real X(*), an array whose length is equal to the
!    column dimension of A'.
!
!    Output, real Y(N), the product A' * X.
!
!    Input, real A(*), integer JA(*), IA(*), the matrix in CSR
!    Compressed Sparse Row format.
!
  implicit none

  integer m,n

  real(kind = dp) a(*)
  integer i
  integer ia(*)
  integer ja(*)
  integer k
  real(kind = dp) x(*)
  real(kind = dp) y(n)

  y = 0.0_dp

  do i = 1, m
      do k = ia(i), ia(i + 1) - 1
          y(ja(k)) = y(ja(k)) + x(i) * a(k)
      enddo
  enddo

  return
end subroutine atmux

! This subroutine is not used in anywhere.
  subroutine csrcsc ( n, job, ipos, a, ja, ia, ao, jao, iao )  
!-------------------------------------------------------------------------------
!  CSRCSC converts Compressed Sparse Row to Compressed Sparse Column.
!
!  Discussion:
!
!    This is essentially a transposition operation.
!
!    It is NOT an in-place algorithm.
!
!    This routine transposes a matrix stored in a, ja, ia format.
!
!  Modified:
!
!    07 January 2004
!
!  Author:
!
!    Youcef Saad
!
!  Parameters:
!
!    Input, integer N, the order of the matrix.
!
!    Input, integer JOB, indicates whether or not to fill the values of the
!    matrix AO or only the pattern (IA, and JA).  Enter 1 for yes.
!
! ipos  = starting position in ao, jao of the transposed matrix.
!         the iao array takes this into account (thus iao(1) is set to ipos.)
!         Note: this may be useful if one needs to append the data structure
!         of the transpose to that of A. In this case use
!                call csrcsc (n,1,n+2,a,ja,ia,a,ja,ia(n+2))
!        for any other normal usage, enter ipos=1.
!
!    Input, real A(*), integer JA(*), IA(N+1), the matrix in CSR
!    Compressed Sparse Row format.
!
!    Output, real AO(*), JAO(*), IAO(N+1), the matrix in CSC
!    Compressed Sparse Column format.
!
  implicit none

  integer n

  real (kind = dp) a(*)
  real (kind = dp) ao(*)
  integer i
  integer ia(n+1)
  integer iao(n+1)
  integer ipos
  integer j
  integer ja(*)
  integer jao(*)
  integer job
  integer k
  integer next
!
!  Compute lengths of rows of A'.
!
  iao(1:n+1) = 0

  do i = 1, n
    do k = ia(i), ia(i+1)-1
      j = ja(k) + 1
      iao(j) = iao(j) + 1
    end do
  end do
!
!  Compute pointers from lengths.
!
  iao(1) = ipos
  do i = 1, n
    iao(i+1) = iao(i) + iao(i+1)
  end do
!
!  Do the actual copying.
!
  do i = 1, n
    do k = ia(i), ia(i+1)-1
      j = ja(k)
      next = iao(j)
      if ( job == 1 ) then
        ao(next) = a(k)
      end if
      jao(next) = i
      iao(j) = next + 1
    end do
  end do
!
!  Reshift IAO and leave.
!
  do i = n, 1, -1
    iao(i+1) = iao(i)
  end do
  iao(1) = ipos

  return
end subroutine csrcsc

! Quick sort for intergers
! Might need to parallelize this for neighbor list construction
  SUBROUTINE qsorti(x,ind,n)
    !***************************************************************************
    !                                                         ROBERT RENKA
    !                                                 OAK RIDGE NATL. LAB.
    !***************************************************************************
    !   THIS SUBROUTINE USES AN ORDER N*LOG(N) QUICK SORT TO SORT AN INTEGER
    ! ARRAY X INTO INCREASING ORDER.  THE ALGORITHM IS AS FOLLOWS.  IND IS
    ! INITIALIZED TO THE ORDERED SEQUENCE OF INDICES 1,...,N, AND ALL INTERCHANGES
    ! ARE APPLIED TO IND.  X IS DIVIDED INTO TWO PORTIONS BY PICKING A CENTRAL
    ! ELEMENT T.  THE FIRST AND LAST ELEMENTS ARE COMPARED WITH T, AND
    ! INTERCHANGES ARE APPLIED AS NECESSARY SO THAT THE THREE VALUES ARE IN
    ! ASCENDING ORDER.  INTERCHANGES ARE THEN APPLIED SO THAT ALL ELEMENTS
    ! GREATER THAN T ARE IN THE UPPER PORTION OF THE ARRAY AND ALL ELEMENTS
    ! LESS THAN T ARE IN THE LOWER PORTION.  THE UPPER AND LOWER INDICES OF ONE
    ! OF THE PORTIONS ARE SAVED IN LOCAL ARRAYS, AND THE PROCESS IS REPEATED
    ! ITERATIVELY ON THE OTHER PORTION.  WHEN A PORTION IS COMPLETELY SORTED,
    ! THE PROCESS BEGINS AGAIN BY RETRIEVING THE INDICES BOUNDING ANOTHER
    ! UNSORTED PORTION.
    !
    ! INPUT PARAMETERS -   N - LENGTH OF THE ARRAY X.
    !
    !                      X - VECTOR OF LENGTH N TO BE SORTED.
    !
    !                    IND - VECTOR OF LENGTH >= N.
    !
    ! N AND X ARE NOT ALTERED BY THIS ROUTINE.
    !
    ! OUTPUT PARAMETER - IND - SEQUENCE OF INDICES 1,...,N PERMUTED IN THE SAME
    !                          FASHION AS X WOULD BE.  THUS, THE ORDERING ON
    !                          X IS DEFINED BY Y(I) = X(IND(I)).
    !
    !*********************************************************************
    !
    ! NOTE -- IU AND IL MUST BE DIMENSIONED >= LOG(N) WHERE LOG HAS BASE 2.
    !
    !*********************************************************************

    ! Code converted using TO_F90 by Alan Miller
    ! Date: 2002-12-18  Time: 11:55:47
    IMPLICIT NONE

    INTEGER, INTENT(IN)  :: x(:)
    INTEGER, INTENT(OUT) :: ind(:)
    INTEGER, INTENT(IN)  :: n


    INTEGER              :: iu(21), il(21)
    INTEGER              :: m, i, j, k, l, ij, it, itt, indx
    REAL(kind=dp)        :: r
    INTEGER              :: t

    ! LOCAL PARAMETERS -

    ! IU,IL =  TEMPORARY STORAGE FOR THE UPPER AND LOWER
    !            INDICES OF PORTIONS OF THE ARRAY X
    ! M =      INDEX FOR IU AND IL
    ! I,J =    LOWER AND UPPER INDICES OF A PORTION OF X
    ! K,L =    INDICES IN THE RANGE I,...,J
    ! IJ =     RANDOMLY CHOSEN INDEX BETWEEN I AND J
    ! IT,ITT = TEMPORARY STORAGE FOR INTERCHANGES IN IND
    ! INDX =   TEMPORARY INDEX FOR X
    ! R =      PSEUDO RANDOM NUMBER FOR GENERATING IJ
    ! T =      CENTRAL ELEMENT OF X

    IF (n <= 0) RETURN

    ! INITIALIZE IND, M, I, J, AND R

    DO  i = 1, n
       ind(i) = i
    END DO
    m = 1
    i = 1
    j = n
    r = .375

    ! TOP OF LOOP

20  IF (i >= j) GO TO 70
    IF (r <= .5898437) THEN
       r = r + .0390625
    ELSE
       r = r - .21875
    END IF

    ! INITIALIZE K

30  k = i
    ! SELECT A CENTRAL ELEMENT OF X AND SAVE IT IN T

    ij = i + r*(j-i)
    it = ind(ij)
    t = x(it)

    ! IF THE FIRST ELEMENT OF THE ARRAY IS GREATER THAN T,
    !   INTERCHANGE IT WITH T

    indx = ind(i)
    IF (x(indx) > t) THEN
       ind(ij) = indx
       ind(i) = it
       it = indx
       t = x(it)
    END IF

    ! INITIALIZE L

    l = j

    ! IF THE LAST ELEMENT OF THE ARRAY IS LESS THAN T,
    !   INTERCHANGE IT WITH T

    indx = ind(j)
    IF (x(indx) >= t) GO TO 50
    ind(ij) = indx
    ind(j) = it
    it = indx
    t = x(it)

    ! IF THE FIRST ELEMENT OF THE ARRAY IS GREATER THAN T,
    !   INTERCHANGE IT WITH T

    indx = ind(i)
    IF (x(indx) <= t) GO TO 50
    ind(ij) = indx
    ind(i) = it
    it = indx
    t = x(it)
    GO TO 50

    ! INTERCHANGE ELEMENTS K AND L
40  itt = ind(l)
    ind(l) = ind(k)
    ind(k) = itt

    ! FIND AN ELEMENT IN THE UPPER PART OF THE ARRAY WHICH IS
    !   NOT LARGER THAN T

50  l = l - 1
    indx = ind(l)
    IF (x(indx) > t) GO TO 50

    ! FIND AN ELEMENT IN THE LOWER PART OF THE ARRAY WHCIH IS NOT SMALLER THAN T

60  k = k + 1
    indx = ind(k)
    IF (x(indx) < t) GO TO 60

    ! IF K <= L, INTERCHANGE ELEMENTS K AND L

    IF (k <= l) GO TO 40

    ! SAVE THE UPPER AND LOWER SUBSCRIPTS OF THE PORTION OF THE
    !   ARRAY YET TO BE SORTED

    IF (l-i > j-k) THEN
       il(m) = i
       iu(m) = l
       i = k
       m = m + 1
       GO TO 80
    END IF

    il(m) = k
    iu(m) = j
    j = l
    m = m + 1
    GO TO 80

    ! BEGIN AGAIN ON ANOTHER UNSORTED PORTION OF THE ARRAY

70  m = m - 1
    IF (m == 0) RETURN
    i = il(m)
    j = iu(m)

80  IF (j-i >= 11) GO TO 30
    IF (i == 1) GO TO 20
    i = i - 1

    ! SORT ELEMENTS I+1,...,J.  NOTE THAT 1 <= I < J AND J-I < 11.

90  i = i + 1
    IF (i == j) GO TO 70
    indx = ind(i+1)
    t = x(indx)
    it = indx
    indx = ind(i)
    IF (x(indx) <= t) GO TO 90
    k = i

100 ind(k+1) = ind(k)
    k = k - 1
    indx = ind(k)
    IF (t < x(indx)) GO TO 100

    ind(k+1) = it
    GO TO 90
  END SUBROUTINE qsorti

end module mod_qclib
