!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**   MODULE mod_stress : Related to the stress tensor
!**
!**   type definitions:
!**   ----------------
!**   N/A
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   avgsize : In virial and hardy stress, avgsize is the radius of the averaging domain
!**             In Tsai stress, avgsize is the length of the surface either upper or lower
!**
!**   Type Variable Definitions:
!**   -------------------------------------
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   N/A
!**
!**   logical Variable Definitions:
!**   ----------------------------
!**   flv : virial stress flag
!**   flt : tsai stress flag
!**   flh : hardy stress flag
!**
!**   Contains Routines:
!**   ------------------
!**   set_stress : Reads in stress related parameters
!**   calculate_stress :
!****************************************************************************
module mod_stress
use mod_global, only : dp,PI
implicit none
save
private avgdomain_hardy_cauchy, &
        avgdomain_virial_cauchy, &
        avgdomain_tsai_cauchy, &
        avgdomain_hardy_pk1, &
        avgdomain_virial_pk1, &
        avgdomain_tsai_pk1, &
        wfunc, &
        wfunc_cauchy, &
        wfunc_pk1, &
        hardy_radius_cauchy, &
        virial_radius_cauchy, &
        tsai_radius_cauchy, &
        hardy_radius_pk1, &
        virial_radius_pk1, &
        tsai_radius_pk1, &
        hardy_volume_cauchy, &
        virial_volume_cauchy, &
        hardy_volume_pk1, &
        virial_volume_pk1, &
        flag_tsai, &
        flag_virial, &
        flag_hardy, &
        flag_cauchy,  &
        flag_pk1,  &
        flag_tsai_cauchy, &
        flag_virial_cauchy, &
        flag_hardy_cauchy, &
        flag_tsai_pk1, &
        flag_virial_pk1, &
        flag_hardy_pk1, &
        characteristic_length_cauchy, &
        characteristic_length_pk1

public  set_stress, &
        compute_stress, &
        deallocate_stress

character(len=512) :: avgdomain_hardy_cauchy = 'sphere',  &
                      avgdomain_virial_cauchy = 'sphere', &
                      avgdomain_tsai_cauchy = 'sphere',   &
                      avgdomain_hardy_pk1 = 'sphere',     &  
                      avgdomain_virial_pk1 = 'sphere',    &
                      avgdomain_tsai_pk1 = 'sphere'

character(len=512) :: wfunc = 'constant',        & ! constant or mollify
                      wfunc_cauchy = 'constant', &
                      wfunc_pk1 = 'constant'

! For spherical average domain
real(kind=dp) :: hardy_radius_cauchy = 1.0_dp, &
                 virial_radius_cauchy = 1.0_dp, &
                 tsai_radius_cauchy = 1.0_dp, &
                 hardy_radius_pk1 = 1.0_dp, &
                 virial_radius_pk1 = 1.0_dp, &
                 tsai_radius_pk1 = 1.0_dp, &
                 hardy_volume_cauchy, &
                 virial_volume_cauchy, &
                 hardy_volume_pk1, &
                 virial_volume_pk1

logical :: flag_tsai            = .false., &
           flag_virial          = .false., &
           flag_hardy           = .false., &
           flag_cauchy          = .true.,  &
           flag_pk1             = .true.,  &
           flag_tsai_cauchy     = .false., &
           flag_virial_cauchy   = .false., &
           flag_hardy_cauchy    = .false., &
           flag_tsai_pk1        = .false., &
           flag_virial_pk1      = .false., &
           flag_hardy_pk1       = .false.

real(kind=dp), dimension(3) :: characteristic_length_cauchy = 1.0_dp, &
                               characteristic_length_pk1    = 1.0_dp

contains

    subroutine set_stress
    use mod_global, only : cmdunit
    use mod_output, only : string,output,erroroutput,nerror,error,heading
    use mod_qclib,  only : next,downcase,freein
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=512) :: input
    character(len=4) :: cc,subcc
    integer :: lower,upper,idum,ierror
    real(kind=dp) :: readin
    logical :: flag = .false.
    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    heading = .true.
    write(string,'(a)') '**Reading stress settings**'; call output
    do
        read(cmdunit,'(a512)',iostat=ierror) input
        if (ierror .ne. 0) then
            nerror = 2
            write(error(1),'(a)') 'Something wrong with the command input.'
            write(error(2),'(a)') 'Macro end must be called to continue the simulation.'
            call erroroutput
        end if

        lower = 0
        upper = next(lower,input)
        cc = input(lower+1:upper-1)

        select case(trim(downcase(cc)))
        case('weig')
            lower = upper
            upper = next(lower,input)
            subcc = input(lower+1:upper-1)
            select case(trim(downcase(subcc)))
                case('cons','')
                    wfunc = 'constant'
                case('moll')
                    wfunc = 'mollify'
                case default
                    nerror = 1
                    error(1) = 'sub macro "' // subcc // '" not recognized in macro weig**'
                    call erroroutput
            end select
            lower = upper
            upper = next(lower,input)
            subcc = input(lower+1:upper-1)
            select case(trim(downcase(subcc)))
                case('cauc')
                    wfunc_cauchy = wfunc
                    if (wfunc == 'constant') then
                        write(string,'(a)') 'Using constant/step weighting function for Cauchy-Hardy stress calculation.'
                        call output
                    else if (wfunc == 'mollify') then
                        write(string,'(a)') 'Using mollifying weighting function for Cauchy-Hardy stress calculation.'
                        call output
                    else
                        nerror = 2
                        write(string,'(2a)') trim(wfunc), ' weighting function is not supported in the code!'
                        write(string,'(a)') 'Program aborted!'
                        call erroroutput
                    end if
                case('pk1')
                    wfunc_pk1 = wfunc
                    if (wfunc == 'constant') then
                        write(string,'(a)') 'Using constant/step weighting function for 1st PK-Hardy stress calculation.'
                        call output
                    else if (wfunc == 'mollify') then
                        write(string,'(a)') 'Using mollifying weighting function for 1st PK-Hardy stress calculation.'
                        call output
                    else
                        nerror = 2
                        write(string,'(2a)') trim(wfunc), ' weighting function is not supported in the code!'
                        write(string,'(a)') 'Program aborted!'
                        call erroroutput
                    end if
                case('')
                    wfunc_cauchy = wfunc
                    wfunc_pk1 = wfunc
                    if (wfunc == 'constant') then
                        write(string,'(a)') &
                        'Using constant/step weighting function for Cauchy-Hardy and 1st pk-Hardy stress calculation.'
                        call output
                    else if (wfunc == 'mollify') then
                        write(string,'(a)') &
                        'Using mollifying weighting function for Cauchy-Hardy and 1st pk-Hardy stress calculation.'
                        call output
                    else
                        nerror = 2
                        write(string,'(2a)') trim(wfunc), ' weighting function is not supported in the code!'
                        write(string,'(a)') 'Program aborted!'
                        call erroroutput
                    end if
                case default
                    nerror = 1
                    error(1) = 'sub macro "' // subcc // '" not recognized in macro weig**'
                    call erroroutput
            end select
        case('avgs')
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,idum,readin,2)
            if (readin <= 0.0_dp) then
                nerror = 2
                write(error(1),'(a)') 'The radius of spherical averaging domain should be > 0.'
                write(error(2),'(a,F25.16)') 'Your input: ', readin
                call erroroutput
            end if

            lower = upper
            upper = next(lower,input)
            subcc = input(lower+1:upper-1)

            select case(trim(downcase(subcc)))
                case('cauc')
                    hardy_radius_cauchy = readin
                    virial_radius_cauchy = readin
                    tsai_radius_cauchy = readin
                    hardy_volume_cauchy = 4.0_dp*PI*(hardy_radius_cauchy**3)/3.0_dp
                    virial_volume_cauchy = 4.0_dp*PI*(virial_radius_cauchy**3)/3.0_dp
                    characteristic_length_cauchy(:) = readin
                    write(string,'(a,3F25.16)') &
                    'Characteristic length for Cauchy stress: ', characteristic_length_cauchy; call output
                    write(string,'(a,F25.16)') &
                    'The radius of the spherical averaging domain for Cauchy stress is: ', readin; call output
                case('pk1')
                    hardy_radius_pk1 = readin
                    virial_radius_pk1 = readin
                    tsai_radius_pk1 = readin
                    hardy_volume_pk1 = 4.0_dp*PI*(hardy_radius_pk1**3)/3.0_dp
                    virial_volume_pk1 = 4.0_dp*PI*(virial_radius_pk1**3)/3.0_dp
                    characteristic_length_pk1(:) = readin
                    write(string,'(a,3F25.16)') &
                    'Characteristic length for 1st pk stress: ', characteristic_length_pk1; call output
                    write(string,'(a,F25.16)') &
                    'The radius of the spherical averaging domain for 1st pk stress is: ', readin; call output
                case('')
                    hardy_radius_cauchy = readin
                    virial_radius_cauchy = readin
                    tsai_radius_cauchy = readin
                    hardy_volume_cauchy = 4.0_dp*PI*(hardy_radius_cauchy**3)/3.0_dp
                    virial_volume_cauchy = 4.0_dp*PI*(virial_radius_cauchy**3)/3.0_dp
                    characteristic_length_cauchy(:) = readin

                    hardy_radius_pk1 = readin
                    virial_radius_pk1 = readin
                    tsai_radius_pk1 = readin
                    hardy_volume_pk1 = 4.0_dp*PI*(hardy_radius_pk1**3)/3.0_dp
                    virial_volume_pk1 = 4.0_dp*PI*(virial_radius_pk1**3)/3.0_dp
                    characteristic_length_pk1(:) = readin
                    write(string,'(a,3F25.16)') &
                    'Characteristic length for Cauchy and 1st pk stress: ', characteristic_length_cauchy; call output
                    write(string,'(a,F25.16)') &
                    'The radius of the spherical averaging domain for Cauchy and 1st pk stress is: ', readin; call output
                case default
                    nerror = 1
                    error(1) = 'sub macro "' // subcc // '" not recognized in macro avgs**'
                    call erroroutput
            end select
        case('cauc')
            lower=upper
            upper=next(lower,input)
            if (downcase(input(lower+1:lower+1)) .eq.'t') then
                flag_cauchy = .true.
            else if (downcase(input(lower+1:lower+1)) .eq.'f') then
                flag_cauchy = .false.
            else
                nerror = 1
                error(1) = 'Unrecognized flag in the Cauchy stress**'
                call erroroutput
            endif
        case('pkst')
            lower=upper
            upper=next(lower,input)
            if (downcase(input(lower+1:lower+1)) .eq.'t') then
                flag_pk1 = .true.
            else if (downcase(input(lower+1:lower+1)) .eq.'f') then
                flag_pk1 = .false.
            else
                nerror = 1
                error(1) = 'Unrecognized flag in the 1st pk stress**'
                call erroroutput
            endif
        case('tsai')
            lower=upper
            upper=next(lower,input)
            if (downcase(input(lower+1:lower+1)) .eq.'t') then
                flag = .true.
                flag_tsai = .true.
            else if (downcase(input(lower+1:lower+1)) .eq.'f') then
                flag = .false.
            else
                nerror = 1
                error(1) = 'Unrecognized flag in the Tsai stress**'
                call erroroutput
            endif
            lower = upper
            upper = next(lower,input)
            subcc = input(lower+1:upper-1)
            select case(trim(downcase(subcc)))
                case('cauc')
                    flag_tsai_cauchy = flag
                    if (flag .and. (flag_cauchy .eqv. .false.)) then
                        flag_cauchy = .true.
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off Cauchy stress.'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'cauchy,true'
                        call output
                    end if
                    write(string,5020) flag_tsai_cauchy; call output
                case('pk1')
                    flag_tsai_pk1 = flag
                    if (flag .and. (flag_pk1 .eqv. .false.)) then
                        flag_pk1 = .true.
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off 1st pk stress.'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'pkstr,true'
                        call output
                    end if
                    write(string,5200) flag_tsai_pk1; call output
                case('')
                    if (flag_cauchy) then
                        flag_tsai_cauchy = flag
                        write(string,5020) flag_tsai_cauchy; call output
                    end if
                    if (flag_pk1) then
                        flag_tsai_pk1 = flag
                        write(string,5200) flag_tsai_pk1; call output
                    end if
                    if (flag .and. (.not. flag_cauchy) .and. (.not. flag_pk1)) then
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off both Cauchy stress and 1st pk stress.'
                        call output
                        write(string,'(a)') &
                        'Assuming you are going to calculate both the Cauchy-Tsai and 1st pk-Tsai stress,'
                        call output
                        write(string,'(a)') &
                        'which will turn both options on, therefore'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'cauchy,true'
                        call output
                        write(string,'(a)') &
                        'pkstr,true'
                        call output
                        flag_cauchy = .true.
                        flag_tsai_cauchy = .true.
                        flag_pk1 = .true.
                        flag_tsai_pk1 = .true.
                        write(string,5020) flag_tsai_cauchy; call output
                        write(string,5200) flag_tsai_pk1; call output
                    end if
                case default
                    nerror = 1
                    error(1) = 'sub macro "' // subcc // '" not recognized in macro tsai**'
                    call erroroutput
            end select
        case('viri')
            lower=upper
            upper=next(lower,input)
            if (downcase(input(lower+1:lower+1)) .eq.'t') then
                flag = .true.
                flag_virial = .true.
            else if (downcase(input(lower+1:lower+1)) .eq.'f') then
                flag = .false.
            else
                nerror = 1
                error(1) = 'Unrecognized flag in the virial stress**'
                call erroroutput
            endif
            lower = upper
            upper = next(lower,input)
            subcc = input(lower+1:upper-1)
            select case(trim(downcase(subcc)))
                case('cauc')
                    flag_virial_cauchy = flag
                    if (flag .and. (flag_cauchy .eqv. .false.)) then
                        flag_cauchy = .true.
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off Cauchy stress.'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'cauchy,true'
                        call output
                    end if
                    write(string,5010) flag_virial_cauchy; call output
                case('pk1')
                    flag_virial_pk1 = flag
                    if (flag .and. (flag_pk1 .eqv. .false.)) then
                        flag_pk1 = .true.
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off 1st pk stress.'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'pkstr,true'
                        call output
                    end if
                    write(string,5100) flag_virial_pk1; call output
                case('')
                    if (flag_cauchy) then
                        flag_virial_cauchy = flag
                        write(string,5010) flag_virial_cauchy; call output
                    end if
                    if (flag_pk1) then
                        flag_virial_pk1 = flag
                        write(string,5100) flag_virial_pk1; call output
                    end if
                    if (flag .and. (.not. flag_cauchy) .and. (.not. flag_pk1)) then
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off both Cauchy stress and 1st pk stress.'
                        call output
                        write(string,'(a)') &
                        'Assuming you are going to calculate both the Cauchy-virial and 1st pk-virial stress,'
                        call output
                        write(string,'(a)') &
                        'which will turn both options on, therefore'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'cauchy,true'
                        call output
                        write(string,'(a)') &
                        'pkstr,true'
                        call output
                        flag_cauchy = .true.
                        flag_virial_cauchy = .true.
                        flag_pk1 = .true.
                        flag_virial_pk1 = .true.
                        write(string,5010) flag_virial_cauchy; call output
                        write(string,5100) flag_virial_pk1; call output
                    end if
                case default
                    nerror = 1
                    error(1) = 'sub macro "' // subcc // '" not recognized in macro virial**'
                    call erroroutput
            end select
        case('hard')
            lower=upper
            upper=next(lower,input)
            if (downcase(input(lower+1:lower+1)) .eq.'t') then
                flag = .true.
                flag_hardy = .true.
            else if (downcase(input(lower+1:lower+1)) .eq.'f') then
                flag = .false.
            else
                nerror = 1
                error(1) = 'Unrecognized flag in the Hardy stress**'
                call erroroutput
            endif
            lower = upper
            upper = next(lower,input)
            subcc = input(lower+1:upper-1)
            select case(trim(downcase(subcc)))
                case('cauc')
                    flag_hardy_cauchy = flag
                    if (flag .and. (flag_cauchy .eqv. .false.)) then
                        flag_cauchy = .true.
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off Cauchy stress.'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'cauchy,true'
                        call output
                    end if
                    write(string,5030) flag_hardy_cauchy; call output
                case('pk1')
                    flag_hardy_pk1 = flag
                    if (flag .and. (flag_pk1 .eqv. .false.)) then
                        flag_pk1 = .true.
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off 1st pk stress.'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'pkstr,true'
                        call output
                    end if
                    write(string,5300) flag_hardy_pk1; call output
                case('')
                    if (flag_cauchy) then
                        flag_hardy_cauchy = flag
                        write(string,5030) flag_hardy_cauchy; call output
                    end if
                    if (flag_pk1) then
                        flag_hardy_pk1 = flag
                        write(string,5300) flag_hardy_pk1; call output
                    end if
                    if (flag .and. (.not. flag_cauchy) .and. (.not. flag_pk1)) then
                        write(string,'(a)') &
                        'Warning: You have a bad input which turns off both Cauchy stress and 1st pk stress.'
                        call output
                        write(string,'(a)') &
                        'Assuming you are going to calculate both the Cauchy-Hardy and 1st pk-Hardy stress,'
                        call output
                        write(string,'(a)') &
                        'which will turn both options on, therefore'
                        call output
                        write(string,'(a)') &
                        'Resetting to:'
                        call output
                        write(string,'(a)') &
                        'cauchy,true'
                        call output
                        write(string,'(a)') &
                        'pkstr,true'
                        call output
                        flag_cauchy = .true.
                        flag_hardy_cauchy = .true.
                        flag_pk1 = .true.
                        flag_hardy_pk1 = .true.
                        write(string,5030) flag_hardy_cauchy; call output
                        write(string,5300) flag_hardy_pk1; call output
                    end if
                case default
                    nerror = 1
                    error(1) = 'sub macro "' // subcc // '" not recognized in macro hardy**'
                    call erroroutput
            end select
        case('end')
            if ((flag_hardy_cauchy .eqv. .false.) .and. &
                (flag_virial_cauchy .eqv. .false.) .and. &
                (flag_tsai_cauchy .eqv. .false.)) then
                flag_cauchy = .false.
            end if

            if ((flag_hardy_pk1 .eqv. .false.) .and. &
                (flag_virial_pk1 .eqv. .false.) .and. &
                (flag_tsai_pk1 .eqv. .false.)) then
                flag_pk1 = .false.
            end if

            if (flag_hardy .or. flag_tsai .or. flag_virial) then
                if (flag_cauchy .or. flag_pk1) then
                    exit
                else
                    nerror = 1
                    write(error(1),'(a)') &
                    'Either Cauchy or 1st PK stress should be specified to calculate! Program terminated!'
                    call erroroutput
                end if
            else
                nerror = 1
                write(error(1),'(a)') 'No stress definition is specified to calculate! Program terminated!'
                call erroroutput
            end if
        case default
            nerror = 1
            error(1) = 'macro "' // cc // '" not recognized in stage stress**'
            call erroroutput
        end select
    enddo

    5010 format('Calculate Cauchy-virial stress      = ',l5)
    5020 format('Calculate Cauchy-Tsai stress        = ',l5)
    5030 format('Calculate Cauchy-Hardy stress       = ',l5)

    5100 format('Calculate 1st pk-virial stress      = ',l5)
    5200 format('Calculate 1st pk-Tsai stress        = ',l5)
    5300 format('Calculate 1st pk-Hardy stress       = ',l5)

    end subroutine set_stress

    subroutine compute_stress
    use mod_forces, only    : compute_forces,rcutoff
    use mod_fij, only       : fij
    use mod_output, only    : string,output,heading
    use mod_box, only       : dimn,nat,samplesize,samplesize_i
    use mod_particles, only : pos,pos_i
    use mod_neigh,only      : form_grid_neigh_list,list,marker1,marker2
    use mod_grid, only      : ngrid,set_grid
    use mod_hardy, only     : hardy,hardy_i,       &
                              virial,virial_i,     &
                              hardy_kin,           &
                              virial_kin,          &
                              hardy_radius,        &
                              hardy_volume,        &
                              virial_radius,       &
                              virial_volume,       &
                              collect_hardy,       &
                              write_hardy_cauchy,  &
                              write_hardy_pk1,     &
                              write_virial_cauchy, &
                              write_virial_pk1
    use mod_tsai, only      : tsai,tsai_i,collect_tsai,tsai_radius,tsai_length,tsai_area,write_tsai_cauchy,write_tsai_pk1
    use mod_rigidity, only  : project,proj_fij
    implicit none

    integer :: i,l
    real(kind=dp) :: tensor(dimn,dimn),tensork(dimn,dimn)

    heading = .true.
    write(string,'(a)') '**Calculating the stress tensor**'; call output

    if (flag_hardy .or. flag_tsai .or. flag_virial) then
        call compute_forces ! NEIGHBOR LIST IS CALCULATED IN THIS SUBROUTINE
    endif

    if (flag_cauchy) then

        heading = .true.
        write(string,'(a)') '**Calculating the Cauchy stress tensor**'; call output

        !-------------------------------------------------------------------------------
        ! Call set_grid to setup the grid on either the reference (pk .eqv. T), or the present
        ! configuration (pk .eqv. F)
        !
        ! TODO:
        call set_grid(characteristic_length_cauchy,'cauchy')

        call form_grid_neigh_list(sqrt(2.0001_dp) * characteristic_length_cauchy + rcutoff)

        if (flag_tsai_cauchy) then
            tsai_radius = tsai_radius_cauchy
            tsai_length = 2.0_dp * tsai_radius_cauchy
            tsai_area = tsai_length ** 2
            write(string,'(a,F25.16)') 'Tsai length Cauchy: ', tsai_length; call output
            write(string,'(a,F25.16)') 'Tsai area Cauchy: ', tsai_area; call output
            if (allocated(tsai)) deallocate(tsai)
            if (.not. allocated(tsai)) allocate(tsai(dimn,dimn,ngrid))
            tsai = 0.0_dp
            if (project) then
                if (allocated(tsai_i)) deallocate(tsai_i)
                if (.not. allocated(tsai_i)) allocate(tsai_i(dimn,dimn,ngrid))
                tsai_i = 0.0_dp
            endif
        endif

        if (flag_hardy_cauchy) then
            hardy_radius = hardy_radius_cauchy
            hardy_volume = hardy_volume_cauchy
            if (allocated(hardy)) deallocate(hardy)
            if (.not. allocated(hardy)) allocate(hardy(dimn,dimn,ngrid))
            if (allocated(hardy_kin)) deallocate(hardy_kin)
            if (.not. allocated(hardy_kin)) allocate(hardy_kin(dimn,dimn,ngrid))
            hardy = 0.0_dp
            hardy_kin = 0.0_dp
            if (project) then
                if (allocated(hardy_i)) deallocate(hardy_i)
                if (.not. allocated(hardy_i)) allocate(hardy_i(dimn,dimn,ngrid))
                hardy_i = 0.0_dp
            endif
        endif

        if (flag_virial_cauchy) then
            virial_radius = virial_radius_cauchy
            virial_volume = virial_volume_cauchy
            if (allocated(virial)) deallocate(virial)
            if (.not. allocated(virial)) allocate(virial(dimn,dimn,ngrid))
            if (allocated(virial_kin)) deallocate(virial_kin)
            if (.not. allocated(virial_kin)) allocate(virial_kin(dimn,dimn,ngrid))
            virial = 0.0_dp
            virial_kin = 0.0_dp
            if (project) then
                if (allocated(virial_i)) deallocate(virial_i)
                if (.not. allocated(virial_i)) allocate(virial_i(dimn,dimn,ngrid))
                virial_i = 0.0_dp
            endif
        endif

        write(string,'(a)') ' '; call output
        write(string,'(a)') 'Looping over all bonds to calculate the Cauchy stress tensors.'; call output

        if (flag_tsai_cauchy) then
            write(string,'(a)') ' '; call output
            write(string,'(a)') 'Looping over all bonds to calculate the Cauchy-Tsai stress tensors.'; call output
            do l = 1, ngrid
                if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                    write(string,'(a,I0)') "grid point Cauchy-Tsai: ", l; call output
                end if
                tensor = 0.0_dp
                call collect_tsai(l,rcutoff,fij,list,marker1,marker2,tensor,&
                                  avgdomain_tsai_cauchy,wfunc_cauchy)
                tsai(:,:,l) = tsai(:,:,l) + tensor
            enddo

            if (project) then
                write(string,'(a)') ' '; call output
                write(string,'(a)') 'Looping over all bonds to calculate the projected Cauchy-Tsai stress tensors.'; call output
                do l = 1, ngrid
                    if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                        write(string,'(a,I0)') "grid point projected Cauchy-Tsai: ", l; call output
                    end if
                    tensor = 0.0_dp
                    call collect_tsai(l,rcutoff,proj_fij,list,marker1,marker2,tensor,&
                                      avgdomain_tsai_cauchy,wfunc_cauchy)
                    tsai_i(:,:,l) = tsai_i(:,:,l) + tensor
                end do
            end if

            call write_tsai_cauchy
        end if

        if (flag_hardy_cauchy) then
            write(string,'(a)') ' '; call output
            write(string,'(a)') 'Looping over all bonds to calculate the Cauchy-Hardy stress tensors.'; call output
            do l = 1, ngrid
                if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                    write(string,'(a,I0)') "grid point Cauchy-Hardy: ", l; call output
                end if
                tensor = 0.0_dp
                tensork = 0.0_dp
                call collect_hardy(l,.true.,rcutoff,fij,list,marker1,marker2,tensor,tensork,&
                                   avgdomain_hardy_cauchy,wfunc_cauchy)
                hardy(:,:,l) = hardy(:,:,l) + tensor
                hardy_kin(:,:,l) = hardy_kin(:,:,l) + tensork
            end do

            if (project) then
                write(string,'(a)') ' '; call output
                write(string,'(a)') 'Looping over all bonds to calculate the projected Cauchy-Hardy stress tensors.'; call output
                do l = 1, ngrid
                    if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                        write(string,'(a,I0)') "grid point projected Cauchy-Hardy: ", l; call output
                    end if
                    tensor = 0.0_dp
                    tensork = 0.0_dp
                    call collect_hardy(l,.true.,rcutoff,proj_fij,list,marker1,marker2,tensor,tensork,&
                                       avgdomain_hardy_cauchy,wfunc_cauchy)
                    hardy_i(:,:,l) = hardy_i(:,:,l) + tensor
                end do
            end if

            call write_hardy_cauchy
        end if

        if (flag_virial_cauchy) then
            write(string,'(a)') ' '; call output
            write(string,'(a)') 'Looping over all bonds to calculate the Cauchy-virial stress tensors.'; call output
            do l = 1, ngrid
                if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                    write(string,'(a,I0)') "grid point Cauchy-virial: ", l; call output
                end if
                tensor = 0.0_dp
                tensork = 0.0_dp
                call collect_hardy(l,.false.,rcutoff,fij,list,marker1,marker2,tensor,tensork,&
                                   avgdomain_virial_cauchy,wfunc_cauchy)
                virial(:,:,l) = virial(:,:,l) + tensor
                virial_kin(:,:,l) = virial_kin(:,:,l) + tensork
            end do

            if (project) then
                write(string,'(a)') ' '; call output
                write(string,'(a)') 'Looping over all bonds to calculate the projected Cauchy-virial stress tensors.'; call output
                do l = 1, ngrid
                    if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                        write(string,'(a,I0)') "grid point projected Cauchy-virial: ", l; call output
                    end if

                    tensor = 0.0_dp
                    tensork = 0.0_dp
                    call collect_hardy(l,.false.,rcutoff,proj_fij,list,marker1,marker2,tensor,tensork,&
                                       avgdomain_virial_cauchy,wfunc_cauchy)
                    virial_i(:,:,l)   = virial_i(:,:,l) + tensor
                end do
            end if

            call write_virial_cauchy
        end if

    end if

    !-------------------------------------------------------------------------------
    ! If P--K stess is calculated then consider the reference configuration
    !
    ! THIS COULD BE DONE IN PARALLEL when distributed

    if (flag_pk1) then

        do i = 1, nat
            pos(:,i) = pos_i(:,i)
        enddo
        samplesize = samplesize_i

        heading = .true.
        write(string,'(a)') '**Calculating the 1st pk stress tensor**'; call output

        !-------------------------------------------------------------------------------
        ! Call set_grid to setup the grid on either the reference (pk .eqv. T), or the present
        ! configuration (pk .eqv. F)
        !
        ! TODO:
        call set_grid(characteristic_length_pk1,'pk1   ')

        call form_grid_neigh_list(sqrt(2.0001_dp) * characteristic_length_pk1 + rcutoff)

        if (flag_tsai_pk1) then
            tsai_radius = tsai_radius_pk1
            tsai_length = 2.0_dp * tsai_radius_pk1
            tsai_area = tsai_length ** 2
            write(string,'(a,F25.16)') 'Tsai length 1st pk: ', tsai_length; call output
            write(string,'(a,F25.16)') 'Tsai area 1st pk: ', tsai_area; call output
            if (allocated(tsai)) deallocate(tsai)
            if (.not. allocated(tsai)) allocate(tsai(dimn,dimn,ngrid))
            tsai = 0.0_dp
            if (project) then
                if (allocated(tsai_i)) deallocate(tsai_i)
                if (.not. allocated(tsai_i)) allocate(tsai_i(dimn,dimn,ngrid))
                tsai_i = 0.0_dp
            endif
        endif

        if (flag_hardy_pk1) then
            hardy_radius = hardy_radius_pk1
            hardy_volume = hardy_volume_pk1
            if (allocated(hardy)) deallocate(hardy)
            if (.not. allocated(hardy)) allocate(hardy(dimn,dimn,ngrid))
            if (allocated(hardy_kin)) deallocate(hardy_kin)
            if (.not. allocated(hardy_kin)) allocate(hardy_kin(dimn,dimn,ngrid))
            hardy = 0.0_dp
            hardy_kin = 0.0_dp
            if (project) then
                if (allocated(hardy_i)) deallocate(hardy_i)
                if (.not. allocated(hardy_i)) allocate(hardy_i(dimn,dimn,ngrid))
                hardy_i = 0.0_dp
            endif
        endif

        if (flag_virial_pk1) then
            virial_radius = virial_radius_pk1
            virial_volume = virial_volume_pk1
            if (allocated(virial)) deallocate(virial)
            if (.not. allocated(virial)) allocate(virial(dimn,dimn,ngrid))
            if (allocated(virial_kin)) deallocate(virial_kin)
            if (.not. allocated(virial_kin)) allocate(virial_kin(dimn,dimn,ngrid))
            virial = 0.0_dp
            virial_kin = 0.0_dp
            if (project) then
                if (allocated(virial_i)) deallocate(virial_i)
                if (.not. allocated(virial_i)) allocate(virial_i(dimn,dimn,ngrid))
                virial_i = 0.0_dp
            endif
        endif

        write(string,'(a)') ' '; call output
        write(string,'(a)') 'Looping over all bonds to calculate the 1st pk stress tensors.'; call output

        if (flag_tsai_pk1) then
            write(string,'(a)') ' '; call output
            write(string,'(a)') 'Looping over all bonds to calculate the 1st pk-Tsai stress tensors.'; call output
            do l = 1, ngrid
                if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                    write(string,'(a,I0)') "grid point 1st pk-Tsai: ", l; call output
                end if
                tensor = 0.0_dp
                call collect_tsai(l,rcutoff,fij,list,marker1,marker2,tensor,&
                                  avgdomain_tsai_pk1,wfunc_pk1)
                tsai(:,:,l) = tsai(:,:,l) + tensor
            end do

            if (project) then
                write(string,'(a)') ' '; call output
                write(string,'(a)') 'Looping over all bonds to calculate the projected 1st pk-Tsai stress tensors.'; call output
                do l = 1, ngrid
                    if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                        write(string,'(a,I0)') "grid point projected 1st pk-Tsai: ", l; call output
                    end if
                    tensor = 0.0_dp
                    call collect_tsai(l,rcutoff,proj_fij,list,marker1,marker2,tensor,&
                                      avgdomain_tsai_pk1,wfunc_pk1)
                    tsai_i(:,:,l) = tsai_i(:,:,l) + tensor
                end do
            end if

            call write_tsai_pk1
        end if

        if (flag_hardy_pk1) then
            write(string,'(a)') ' '; call output
            write(string,'(a)') 'Looping over all bonds to calculate the 1st pk-Hardy stress tensors.'; call output
            do l = 1, ngrid
                if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                    write(string,'(a,I0)') "grid point 1st pk-Hardy: ", l; call output
                end if
                tensor = 0.0_dp
                tensork = 0.0_dp
                call collect_hardy(l,.true.,rcutoff,fij,list,marker1,marker2,tensor,tensork,&
                                   avgdomain_hardy_pk1,wfunc_pk1)
                hardy(:,:,l) = hardy(:,:,l) + tensor
                hardy_kin(:,:,l) = hardy_kin(:,:,l) + tensork
            end do

            if (project) then
                write(string,'(a)') ' '; call output
                write(string,'(a)') 'Looping over all bonds to calculate the 1st pk-Hardy stress tensors.'; call output
                do l = 1, ngrid
                    if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                        write(string,'(a,I0)') "grid point projected 1st pk-Hardy: ", l; call output
                    end if
                    tensor = 0.0_dp
                    tensork = 0.0_dp
                    call collect_hardy(l,.true.,rcutoff,proj_fij,list,marker1,marker2,tensor,tensork,&
                                       avgdomain_hardy_pk1,wfunc_pk1)
                    hardy_i(:,:,l) = hardy_i(:,:,l) + tensor
                end do
            end if

            call write_hardy_pk1
        end if

        if (flag_virial_pk1) then
            write(string,'(a)') ' '; call output
            write(string,'(a)') 'Looping over all bonds to calculate the 1st pk-virial stress tensors.'; call output
            do l = 1, ngrid
                if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                    write(string,'(a,I0)') "grid point 1st pk-virial: ", l; call output
                end if
                tensor = 0.0_dp
                tensork = 0.0_dp
                call collect_hardy(l,.false.,rcutoff,fij,list,marker1,marker2,tensor,tensork,&
                                   avgdomain_virial_pk1,wfunc_pk1)
                virial(:,:,l) = virial(:,:,l) + tensor
                virial_kin(:,:,l) = virial_kin(:,:,l) + tensork
            enddo

            if (project) then
                write(string,'(a)') ' '; call output
                write(string,'(a)') 'Looping over all bonds to calculate the 1st pk-virial stress tensors.'; call output
                do l = 1, ngrid
                    if ((mod(l, 100) .eq. 0) .or. (l == ngrid)) then
                        write(string,'(a,I0)') "grid point projected 1st pk-virial: ", l; call output
                    end if
                    tensor = 0.0_dp
                    tensork = 0.0_dp
                    call collect_hardy(l,.false.,rcutoff,proj_fij,list,marker1,marker2,tensor,tensork,&
                                       avgdomain_virial_pk1,wfunc_pk1)
                    virial_i(:,:,l) = virial_i(:,:,l) + tensor
                end do
            end if

            call write_virial_pk1
        end if

    end if

    write(string,'(a)') 'Stress computation finished!'; call output

    end subroutine compute_stress

    subroutine deallocate_stress
    use mod_hardy, only : hardy, hardy_i,   &
                          virial, virial_i, &
                          hardy_kin,        &
                          virial_kin
    use mod_tsai, only  : tsai,tsai_i
    implicit none
    if (allocated(hardy)) deallocate(hardy)
    if (allocated(hardy_i)) deallocate(hardy_i)
    if (allocated(hardy_kin)) deallocate(hardy_kin)
    if (allocated(virial)) deallocate(virial)
    if (allocated(virial_i)) deallocate(virial_i)
    if (allocated(virial_kin)) deallocate(virial_kin)
    if (allocated(tsai)) deallocate(tsai)
    if (allocated(tsai_i)) deallocate(tsai_i)
    end subroutine deallocate_stress

end module mod_stress
